"""Fulfillment routes processing methods and functions."""

# Copyright 2023 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Dict, Any

from dfcx_scrapi.agent_extract import common
from dfcx_scrapi.agent_extract import types


class Fulfillments:
    """Fulfillment routes processing methods and functions."""

    def __init__(self):
        self.common = common.Common()
        self.route_parameters = {}

    @staticmethod
    def check_for_webhook(page: types.Page, path: Dict[str, Any]):
        """Check the current route for existence of webhook."""
        if "webhook" in path:
            page.has_webhook = True

    @staticmethod
    def check_for_webhook_event_handlers(route: types.Fulfillment):
        """Check for Webhook Error Event Handler on Page.

        In this method, we're interested in the following conditions:
         - Page is currently flagged w/webhook = True
         - Page HAS NOT been flagged w/having a webhook error handler
         - The trigger MATCHES the pattern 'webhook.error'

        If a Page and its Route meet all the criteria, we'll flip the bit.
        Otherwise, the webhook handler bit will remain False, causing a rule
        flag."""

        if all(
            [
                route.page.has_webhook,
                not route.page.has_webhook_event_handler,
                "webhook.error" in route.trigger,
            ]
        ):
            route.page.has_webhook_event_handler = True

    @staticmethod
    def check_for_intent(route: types.Fulfillment):
        """Check route data to see if Intent is present."""
        intent = None
        if "intent" in route.data:
            intent = route.data["intent"]

        return intent

    def process_intents_in_routes(
            self, route: types.Fulfillment, stats: types.AgentData):
        intent = self.check_for_intent(route)
        if intent:
            pair = (intent, route.page.display_name)
            stats.active_intents[
                route.page.flow.display_name].append(pair)

        return stats

    def collect_transition_route_trigger(self, route):
        """Inspect route and return all Intent/Condition info."""

        trigger = []

        if "intent" in route.data:
            trigger.append("intent")

        if "condition" in route.data:
            trigger.append("condition")

        if len(trigger) > 0:
            trigger = "+".join(trigger)

        return trigger

    def get_trigger_info(self, route):
        """Extract trigger info from route based on primary key."""

        if route.fulfillment_type == "event":
            trigger = f"event : {route.data.get('event', None)}"

        if route.fulfillment_type == "reprompt_handler":
            trigger = f"{route.parameter} : event : "\
                f"{route.data.get('event', None)}"

        if route.fulfillment_type == "transition_route":
            intent_condition = self.collect_transition_route_trigger(route)
            trigger = f"route : {intent_condition}"

        return trigger

    def set_route_group_targets(self, page: types.Page, stats: types.AgentData):
        """Determine Route Targets for Route Group routes."""
        current_page = page.display_name

        for route_group in page.route_groups:
            page.flow.graph.add_edge(current_page, route_group)
            page.flow.graph.add_used_node(route_group)

            stats.graph.add_edge(current_page, route_group)
            stats.graph.add_used_node(route_group)

        return page, stats

    def set_route_targets(
            self, route: types.Fulfillment, stats: types.AgentData):
        """Determine the Route Targets for the specified route.

        Primary function is to build out the graph structure for the
        Flow based on the current page and where the routes are pointing to.
        The graph structure can then be traversed later to determine any errors
        or inconsistencies in design.
        """
        current_page = route.page.display_name

        route.target_flow = route.data.get("targetFlow", None)
        route.target_page = route.data.get("targetPage", None)

        if route.target_page:
            route.page.flow.graph.add_edge(current_page, route.target_page)
            route.page.flow.graph.add_used_node(route.target_page)

            stats.graph.add_edge(current_page, route.target_page)
            stats.graph.add_used_node(route.target_page)

        if route.target_flow:
            route.page.flow.graph.add_edge(
                current_page, f"FLOW: {route.target_flow}")
            route.page.flow.graph.add_used_node(f"FLOW: {route.target_flow}")

            stats.graph.add_edge(
                current_page, f"FLOW: {route.target_flow}"
            )
            stats.graph.add_used_node(f"FLOW: {route.target_flow}")

        return route, stats

    def update_route_parameters(
            self, route: types.Fulfillment, item: Dict[str, str]):
        """Update the Route Parameters map based on new info."""
        flow_name = route.page.flow.display_name
        page_name = route.page.display_name

        flow_data = self.route_parameters.get(flow_name, None)
        page_data = None

        if flow_data:
            page_data = flow_data.get(page_name, None)

        # Flow and Page already exists, append to existing list.
        if page_data:
            self.route_parameters[flow_name][page_name].append(item)

        # Flow data exists, but not Page, so only create the Page list.
        elif flow_data and not page_data:
            self.route_parameters[flow_name][page_name] = [item]

        # Neither the Flow or Page data exists, so create it all.
        else:
            self.route_parameters[flow_name] = {page_name: [item]}


    def process_fulfillment_type(
        self, stats: types.AgentData, route: types.Fulfillment, path: object,
        key: str):
        """Parse through specific fulfillment types."""
        fulfillment_data = path.get(key, None)

        if fulfillment_data:
            for item in fulfillment_data:
                # This is where each message type will exist
                # text, custom payload, etc.

                if "text" in item:
                    for text in item["text"]["text"]:
                        route.text = text

                if "parameter" in item:
                    self.update_route_parameters(route, item)

        return stats

    def process_reprompt_handlers(
            self, fp: types.FormParameter, stats: types.AgentData):
        """Processing for Reprompt Event Handlers inside Form parameters.

        While Reprompt Event Handlers are technically Events, they differ from
        standard Page level Events because they act on the FormParameter data
        structure, not Fulfillment Route data structure as standard Events do.
        """
        if not fp.reprompt_handlers:
            return stats

        for handler in fp.reprompt_handlers:
            route = types.Fulfillment(page=fp.page)
            route.data = handler
            route.agent_id = fp.page.agent_id
            route.fulfillment_type = "reprompt_handler"
            route.parameter = fp.display_name
            route.trigger = self.get_trigger_info(route)
            route, stats = self.set_route_targets(route, stats)
            path = route.data.get("triggerFulfillment", None)
            event = route.data.get("event", None)

            stats = self.process_intents_in_routes(route, stats)

            if not path and not event:
                continue

            # Flag for Webhook Handler
            self.check_for_webhook(fp.page, path)

            stats = self.process_fulfillment_type(
                stats, route, path, "messages")

        return stats

    def process_events(self, page: types.Page, stats: types.AgentData):
        """Parse through all Page Event Handlers."""
        if not page.events:
            return stats

        for route_data in page.events:
            route = types.Fulfillment(page=page)
            route.data = route_data
            route.agent_id = page.agent_id
            route.fulfillment_type = "event"
            route.trigger = self.get_trigger_info(route)
            route, stats = self.set_route_targets(route, stats)
            path = route.data.get("triggerFulfillment", None)
            event = route.data.get("event", None)

            stats = self.process_intents_in_routes(route, stats)

            if not path and not event:
                continue

            # Flag for Webhook Handler
            self.check_for_webhook_event_handlers(route)

            stats = self.process_fulfillment_type(
                stats, route, path, "messages")

        return stats

    def process_routes(self, page: types.Page, stats: types.AgentData):
        """Parse through all Transition Routes."""
        tf_key = "triggerFulfillment"

        if not page.routes:
            return stats

        for route_data in page.routes:
            route = types.Fulfillment(page=page)
            route.data = route_data
            route.agent_id = page.agent_id
            route.fulfillment_type = "transition_route"
            route.trigger = self.get_trigger_info(route)
            route, stats = self.set_route_targets(route, stats)

            stats = self.process_intents_in_routes(route, stats)

            path = route.data.get(tf_key, None)

            if not path:
                continue

            # Flag for Webhook Handler
            self.check_for_webhook(page, path)

            stats = self.process_fulfillment_type(
                stats, route, path, "messages")

            # Preset Params processed here
            stats = self.process_fulfillment_type(
                stats, route, path, "setParameterActions"
            )

        return stats

    def process_entry(self, page: types.Page, stats: types.AgentData):
        """Process Entry Fulfillment on a single page file.

        The Entry Fulfillment to a Page only has 1 "route" (i.e. itself) so
        there is no need to loop through multiple routes, as they don't
        exist for Entry Fulfillment.
        """

        if not page.entry:
            return stats

        route = types.Fulfillment(page=page)
        route.data = page.entry
        route.agent_id = page.agent_id
        route.fulfillment_type = "entry"
        route.trigger = "entry"
        path = route.data

        self.check_for_webhook(page, path)

        stats = self.process_fulfillment_type(stats, route, path, "messages")

        return stats



"""A set of builder methods to create CX proto resource objects"""

# Copyright 2023 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging
from dataclasses import dataclass
from typing import List, Union

from google.cloud.dialogflowcx_v3beta1.types import Flow
from google.cloud.dialogflowcx_v3beta1.types import NluSettings
from google.cloud.dialogflowcx_v3beta1.types import TransitionRoute
from google.cloud.dialogflowcx_v3beta1.types import EventHandler
from dfcx_scrapi.builders.builders_common import BuildersCommon
from dfcx_scrapi.builders.routes import TransitionRouteBuilder
from dfcx_scrapi.builders.routes import EventHandlerBuilder
from dfcx_scrapi.builders.fulfillments import FulfillmentBuilder

# logging config
logging.basicConfig(
    level=logging.INFO,
    format="%(asctime)s %(levelname)-8s %(message)s",
    datefmt="%Y-%m-%d %H:%M:%S",
)


class FlowBuilder(BuildersCommon):
    """Base Class for CX Flow builder."""

    _proto_type = Flow
    _proto_type_str = "Flow"


    def __str__(self) -> str:
        """String representation of the proto_obj."""
        self._check_proto_obj_attr_exist()

        return (
            f"Basic Information:\n{'='*25}\n{self._show_basic_info()}"
            f"\n\n\nTransitionRoutes:\n{'='*25}"
            f"\n{self._show_transition_routes()}"
            f"\n\n\nEventHandlers:\n{'='*25}\n{self._show_event_handlers()}"
            f"\n\n\nTransitoinRouteGroups:\n{'='*25}"
            f"\n{self._show_transition_route_groups()}")


    def _show_basic_info(self) -> str:
        """String representation for the basic information of proto_obj."""
        self._check_proto_obj_attr_exist()

        nlu_settings_str = (
            f"\tModel type: {self.proto_obj.nlu_settings.model_type.name}"
            "\n\tClassification threshold:"
            f" {self.proto_obj.nlu_settings.classification_threshold}"
            "\n\tTraining mode:"
            f" {self.proto_obj.nlu_settings.model_training_mode.name}"
        )

        return (
            f"display_name: {self.proto_obj.display_name}"
            f"\ndescription:\n\t{self.proto_obj.description}"
            f"\nNLU settings:\n{nlu_settings_str}"
        )


    def _show_transition_routes(self) -> str:
        """String representation for the transition routes of proto_obj."""
        self._check_proto_obj_attr_exist()

        return "\n".join([
            f"TransitionRoute {i+1}:\n{str(TransitionRouteBuilder(tr))}"
            f"\n{'*'*20}\n"
            for i, tr in enumerate(self.proto_obj.transition_routes)
        ])


    def _show_event_handlers(self) -> str:
        """String representation for the event handlers of proto_obj."""
        self._check_proto_obj_attr_exist()

        return "\n".join([
            f"EventHandler {i+1}:\n{str(EventHandlerBuilder(eh))}\n{'*'*20}\n"
            for i, eh in enumerate(self.proto_obj.event_handlers)
        ])


    def _show_transition_route_groups(self) -> str:
        """String representation for the transition route groups of proto_obj"""
        self._check_proto_obj_attr_exist()

        return "\n".join([
            f"TransitionRouteGroup {i+1}: {trg_id}"
            for i, trg_id in enumerate(self.proto_obj.transition_route_groups)
        ])


    def show_flow_info(
        self, mode: str = "whole"
    ) -> None:
        """Show the proto_obj information.

        Args:
          mode (str):
            Specifies what part of the page to show.
              Options:
              ['basic', 'whole',
              'routes' or 'transition routes',
              'route groups' or 'transition route groups',
              'events' or 'event handlers'
              ]
        """
        self._check_proto_obj_attr_exist()

        if mode == "basic":
            print(self._show_basic_info())
        elif mode in ["routes", "transition routes"]:
            print(self._show_transition_routes())
        elif mode in ["route groups", "transition route groups"]:
            print(self._show_transition_route_groups())
        elif mode in ["events", "event handlers"]:
            print(self._show_event_handlers())
        elif mode == "whole":
            print(self)
        else:
            raise ValueError(
                "mode should be in"
                "['basic', 'whole',"
                " 'routes', 'transition routes',"
                " 'route groups', 'transition route groups',"
                " 'events', 'event handlers']"
            )


    def show_stats(self) -> None:
        """Provide some stats about the Page."""
        self._check_proto_obj_attr_exist()

        stats_instance = FlowStats(self.proto_obj)
        stats_instance.generate_stats()


    def create_new_proto_obj(
        self,
        display_name: str,
        description: str = None,
        overwrite: bool = False
    ) -> Flow:
        """Create a new Flow.

        Args:
          display_name (str):
            Required. The human-readable name of the flow.
          description (str):
            The description of the flow. The maximum length is 500 characters.
            If exceeded, the request is rejected.
          overwrite (bool)
            Overwrite the new proto_obj if proto_obj already contains a Flow.

        Returns:
          A Flow object stored in proto_obj.
        """
        # Types error checking
        if not(display_name and isinstance(display_name, str)):
            raise ValueError("`display_name` should be a nonempty string.")
        if (
            description
            and not isinstance(description, str)
            and len(description) > 500
        ):
            raise ValueError(
                "`description` should be a string and"
                " it's length should be less than 500 characters."
            )
        # `overwrite` parameter error checking
        if self.proto_obj and not overwrite:
            raise UserWarning(
                "proto_obj already contains a Flow."
                " If you wish to overwrite it, pass `overwrite` as True."
            )
        # Create the Flow
        if overwrite or not self.proto_obj:
            self.proto_obj = Flow(
                display_name=display_name,
                description=description
            )
        # Set the NLU settings to default
        self.nlu_settings()

        return self.proto_obj


    def nlu_settings(
        self,
        model_type: int = 1,
        classification_threshold: float = 0.3,
        model_training_mode: int = 1
    ) -> Flow:
        """NLU related settings of the flow.

        Args:
          model_type (int):
            Indicates the type of NLU model:
              1 = MODEL_TYPE_STANDARD, 3 = MODEL_TYPE_ADVANCED
          classification_threshold (float):
            To filter out false positive results and
            still get variety in matched natural language
            inputs for your agent, you can tune the machine
            learning classification threshold. If the
            returned score value is less than the threshold
            value, then a no-match event will be triggered.
            The score values range from 0.0 (completely
            uncertain) to 1.0 (completely certain). If set
            to 0.0, the default of 0.3 is used.
          model_training_mode (int):
            Indicates NLU model training mode:
              1 = MODEL_TRAINING_MODE_AUTOMATIC
              2 = MODEL_TRAINING_MODE_MANUAL

        Returns:
          A Flow object stored in proto_obj.
        """
        self._check_proto_obj_attr_exist()

        # Type error checking
        if model_type not in [1, 3]:
            raise ValueError(
                "`model_type` should be in [1, 3]."
                "\n1: MODEL_TYPE_STANDARD"
                "\n3: MODEL_TYPE_ADVANCED"
            )
        if model_training_mode not in [1, 2]:
            raise ValueError(
                "`model_training_mode` should be in [1, 2]."
                "\n1: MODEL_TRAINING_MODE_AUTOMATIC"
                "\n2: MODEL_TRAINING_MODE_MANUAL"
            )
        if not(
            isinstance(classification_threshold, float)
            and (0 < classification_threshold < 1)
        ):
            raise ValueError(
                "`classification_threshold` should be a float"
                " range from 0.0 to 1.0."
            )

        the_nlu_settings = NluSettings(
            model_type=model_type,
            model_training_mode=model_training_mode,
            classification_threshold=classification_threshold
        )
        self.proto_obj.nlu_settings = the_nlu_settings

        return self.proto_obj


    def add_transition_route(
        self,
        transition_routes: Union[TransitionRoute, List[TransitionRoute]]
    ) -> Flow:
        """Add single or multiple TransitionRoutes to the Flow.

        Args:
          transition_routes (TransitionRoute | List[TransitionRoute]):
            A single or list of TransitionRoutes to add
            to the Flow existing in proto_obj.

        Returns:
          A Flow object stored in proto_obj.
        """
        self._check_proto_obj_attr_exist()

        # Type error checking
        self._is_type_or_list_of_types(
            transition_routes, TransitionRoute, "transition_routes"
        )

        if not isinstance(transition_routes, list):
            transition_routes = [transition_routes]
        self.proto_obj.transition_routes.extend(transition_routes)

        return self.proto_obj


    def add_event_handler(
        self,
        event_handlers: Union[EventHandler, List[EventHandler]]
    ) -> Flow:
        """Add single or multiple EventHandlers to the Flow.

        Args:
          event_handlers (EventHandler | List[EventHandler]):
            A single or list of EventHandler to add
            to the Flow existing in proto_obj.

        Returns:
          A Flow object stored in proto_obj.
        """
        self._check_proto_obj_attr_exist()

        # Type error checking
        self._is_type_or_list_of_types(
            event_handlers, EventHandler, "event_handlers"
        )

        if not isinstance(event_handlers, list):
            event_handlers = [event_handlers]
        self.proto_obj.event_handlers.extend(event_handlers)

        return self.proto_obj


    def add_transition_route_group(
        self,
        transition_route_groups: Union[str, List[str]]
    ) -> Flow:
        """Add single or multiple TransitionRouteGroups to the Flow.

        Args:
          transition_route_groups (str | List[str]):
            A single or list of TransitionRouteGroup's id to add
            to the Flow existed in proto_obj. Format:
            ``projects/<Project ID>/locations/<Location ID>/agents/<Agent ID>/
              flows/<Flow ID>/transitionRouteGroups/<TransitionRouteGroup ID>``.

        Returns:
          A Flow object stored in proto_obj.
        """
        self._check_proto_obj_attr_exist()

        # Type error checking
        self._is_type_or_list_of_types(
            transition_route_groups, str, "transition_route_groups"
        )

        if not isinstance(transition_route_groups, list):
            transition_route_groups = [transition_route_groups]
        self.proto_obj.transition_route_groups.extend(transition_route_groups)

        return self.proto_obj


    def remove_transition_route(
        self,
        transition_route: TransitionRoute = None,
        intent: str = None,
        condition: str = None
    ) -> Flow:
        """Remove a transition route from the Flow.

        At least one of the `transition_route`, `intent`, or `condition` should
        be specfied.

        Args:
            transition_route (TransitionRoute):
              The TransitionRoute to remove from the Flow.
            intent (str):
              TransitionRoute's intent that should be removed from the Flow.
            condition (str):
              TransitionRoute's condition that should be removed from the Flow.

        Returns:
          A Flow object stored in proto_obj.
        """
        self._check_proto_obj_attr_exist()

        new_routes = []
        for tr in self.proto_obj.transition_routes:
            if self._match_transition_route(
                transition_route=tr, target_route=transition_route,
                intent=intent, condition=condition
            ):
                continue
            new_routes.append(tr)
        self.proto_obj.transition_routes = new_routes

        return self.proto_obj


    def remove_event_handler(
        self,
        event_handlers: Union[EventHandler, List[EventHandler]] = None,
        event_names: Union[str, List[str]] = None
    ) -> Flow:
        """Remove single or multiple EventHandlers from the Flow.

        Args:
          event_handlers (EventHandler | List[EventHandler]):
            A single or list of EventHandler to remove
              from the Flow existing in proto_obj.
            Only one of the `event_handlers` and
              `event_names` should be specified.
          event_names (str | List[str]):
            A single or list of EventHandler's event names corresponding to the
              EventHandler to remove from the Flow existing in proto_obj.
            Only one of the `event_handlers` and
              `event_names` should be specified.

        Returns:
          A Flow object stored in proto_obj.
        """
        self._check_proto_obj_attr_exist()

        if event_handlers and event_names:
            raise UserWarning(
                "Only one of the `event_handlers` and "
                "`event_names` should be specified."
            )
        if event_handlers:
            new_ehs = self._find_unmatched_event_handlers(event_handlers)
        elif event_names:
            new_ehs = self._find_unmatched_event_handlers_by_name(event_names)
        else:
            raise UserWarning(
                "At least one of the `event_handlers` and "
                "`event_names` should be specified."
            )

        self.proto_obj.event_handlers = new_ehs

        return self.proto_obj


    def remove_transition_route_group(
        self,
        transition_route_groups: Union[str, List[str]]
    ) -> Flow:
        """Remove single or multiple TransitionRouteGroups from the Flow.

        Args:
          transition_route_groups (str | List[str]):
            A single or list of TransitionRouteGroup's id to remove
            from the Flow existing in proto_obj. Format:
            ``projects/<Project ID>/locations/<Location ID>/agents/<Agent ID>/
              flows/<Flow ID>/transitionRouteGroups/<TransitionRouteGroup ID>``.

        Returns:
          A Flow object stored in proto_obj.
        """
        self._check_proto_obj_attr_exist()

        # Type error checking
        self._is_type_or_list_of_types(
            transition_route_groups, str, "transition_route_groups"
        )

        if not isinstance(transition_route_groups, list):
            transition_route_groups = [transition_route_groups]

        new_trgs = [
            trg
            for trg in self.proto_obj.transition_route_groups
            if trg not in transition_route_groups
        ]
        self.proto_obj.transition_route_groups = new_trgs

        return self.proto_obj



@dataclass
class FlowStats():
    """A class for tracking the stats of CX Flow object."""
    flow_proto_obj: Flow

    # Transition Routes
    transition_routes_count: int = 0
    routes_with_fulfill_count: int = 0
    routes_with_webhook_fulfill_count: int = 0
    intent_routes_count: int = 0
    cond_routes_count: int = 0
    intent_and_cond_routes_count: int = 0

    # Event Handlers
    event_handlers_count: int = 0
    events_with_fulfill_count: int = 0
    events_with_webhook_fulfill_count: int = 0

    # Transition Route Groups
    transition_route_groups_count: int = 0


    def calc_transition_route_stats(self):
        """Calculating TransitionRoute related stats"""
        self.transition_routes_count = len(
            self.flow_proto_obj.transition_routes
        )
        for tr in self.flow_proto_obj.transition_routes:
            if tr.trigger_fulfillment:
                self.routes_with_fulfill_count += 1
                fb = FulfillmentBuilder(tr.trigger_fulfillment)
                if fb.has_webhook():
                    self.routes_with_webhook_fulfill_count += 1
            if tr.intent and tr.condition:
                self.intent_and_cond_routes_count += 1
            elif tr.intent and not tr.condition:
                self.intent_routes_count += 1
            elif not tr.intent and tr.condition:
                self.cond_routes_count += 1

    def create_transition_route_str(self) -> str:
        """String representation of TransitionRoutes stats."""
        transition_routes_str = (
            f"# of Transition Routes: {self.transition_routes_count}"
        )
        routes_with_fulfill_str = (
            f"# of routes with fulfillment: {self.routes_with_fulfill_count}"
        )
        routes_with_webhook_fulfill_str = (
            "# of routes uses webhook for fulfillment:"
            f" {self.routes_with_webhook_fulfill_count}"
        )
        intent_routes_str = f"# of intent routes: {self.intent_routes_count}"
        cond_routes_str = f"# of condition routes: {self.cond_routes_count}"
        intent_and_cond_routes_str = (
            "# of intent and condition routes:"
            f" {self.intent_and_cond_routes_count}"
        )

        return (
            f"{transition_routes_str}\n\t{intent_routes_str}"
            f"\n\t{cond_routes_str}\n\t{intent_and_cond_routes_str}"
            f"\n\t{routes_with_fulfill_str}"
            f"\n\t{routes_with_webhook_fulfill_str}"
        )


    def calc_event_handler_stats(self):
        """Calculating EventHandler related stats."""
        self.event_handlers_count = len(self.flow_proto_obj.event_handlers)
        for eh in self.flow_proto_obj.event_handlers:
            fb = FulfillmentBuilder(eh.trigger_fulfillment)
            if fb.has_webhook():
                self.events_with_webhook_fulfill_count += 1
            if eh.trigger_fulfillment:
                self.events_with_fulfill_count += 1

    def create_event_handler_str(self) -> str:
        """String representation of EventHandlers stats."""
        event_handlers_str = f"# of Event Handlers: {self.event_handlers_count}"
        events_with_fulfill_str = (
            "# of Event Handlers with fulfillment:"
            f" {self.events_with_fulfill_count}"
        )
        events_with_webhook_fulfill_str = (
            "# of Event Handlers uses webhook for fulfillment:"
            f" {self.events_with_webhook_fulfill_count}"
        )

        return (
            f"{event_handlers_str}\n\t{events_with_fulfill_str}"
            f"\n\t{events_with_webhook_fulfill_str}"
        )


    def create_transition_route_group_str(self) -> str:
        """String representation of TransitionRouteGroup stats."""
        self.transition_route_groups_count = len(
            self.flow_proto_obj.transition_route_groups
        )
        return (
            "# of Transition Route Groups:"
            f" {self.transition_route_groups_count}"
        )


    def generate_stats(self):
        """Generate stats for the Flow."""
        self.calc_transition_route_stats()
        self.calc_event_handler_stats()

        routes_stats_str = self.create_transition_route_str()
        events_stats_str = self.create_event_handler_str()
        route_groups_stats_str = self.create_transition_route_group_str()

        out = (
            f"{routes_stats_str}\n{events_stats_str}\n{route_groups_stats_str}"
        )
        print(out)



"""A set of builder methods to create CX proto resource objects"""

# Copyright 2023 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging
from typing import List, Union

from google.cloud.dialogflowcx_v3beta1.types import EntityType
from dfcx_scrapi.builders.builders_common import BuildersCommon

# logging config
logging.basicConfig(
    level=logging.INFO,
    format="%(asctime)s %(levelname)-8s %(message)s",
    datefmt="%Y-%m-%d %H:%M:%S",
)


class EntityTypeBuilder(BuildersCommon):
    """Base Class for CX EntityType builder."""

    _proto_type = EntityType
    _proto_type_str = "EntityType"


    def __str__(self) -> str:
        """String representation of the proto_obj."""
        self._check_proto_obj_attr_exist()

        return (
            f"{self._show_entity_type_basic_info()}"
            f"\n{self._show_excluded_phrases()}"
            f"\n{self._show_entities()}")


    def _show_entity_type_basic_info(self) -> str:
        """Shows the information of proto_obj."""
        self._check_proto_obj_attr_exist()

        return (
            f"display_name: {self.proto_obj.display_name}"
            f"\nkind: {self.proto_obj.kind.name}"
            f"\nauto_expansion_mode: {self.proto_obj.auto_expansion_mode}"
            "\nenable_fuzzy_extraction:"
            f" {self.proto_obj.enable_fuzzy_extraction}"
            f"\nredact: {self.proto_obj.redact}")

    def _show_excluded_phrases(self) -> str:
        """Shows the excluded phrases of proto_obj."""
        self._check_proto_obj_attr_exist()

        excluded_phrases = "\n\t".join([
            phrase.value
            for phrase in self.proto_obj.excluded_phrases
        ])

        return f"excluded phrases:\n\t{excluded_phrases}"

    def _show_entities(self) -> str:
        """Shows the entities of proto_obj."""
        self._check_proto_obj_attr_exist()

        entities =  "\n".join([
            f"  {entity.value}:\n\t{', '.join(entity.synonyms)}"
            for entity in self.proto_obj.entities
        ])

        return f"entities:\n{entities}"


    def show_entity_type(self, mode: str = "whole"):
        """Show the proto_obj information.

        Args:
          mode (str):
            Specifies what part of the entity type to show.
            Options:
              ['basic', 'entities', 'excluded' or 'excluded phrases', 'whole']
        """
        self._check_proto_obj_attr_exist()

        if mode == "basic":
            print(self._show_entity_type_basic_info())
        elif mode == "entities":
            print(self._show_entities())
        elif mode in ["excluded", "excluded phrases"]:
            print(self._show_excluded_phrases())
        elif mode == "whole":
            print(self)
        else:
            raise ValueError(
                "mode should be in"
                "['basic', 'entities',"
                " 'excluded' or 'excluded phrases', 'whole']"
            )


    def create_new_proto_obj(
        self,
        display_name: str,
        kind: int,
        auto_expansion_mode: int = 0,
        enable_fuzzy_extraction: bool = False,
        redact: bool = False,
        overwrite: bool = False
    ) -> EntityType:
        """Create a new EntityType.

        Args:
          display_name (str):
            The human-readable name of the
            entity type, unique within the agent.
          kind (int):
            Represents kinds of entities.
              1 = KIND_MAP
              2 = KIND_LIST
              3 = KIND_REGEXP
          auto_expansion_mode (int):
            Indicates whether the entity type can be
            automatically expanded.
              AUTO_EXPANSION_MODE_UNSPECIFIED = 0
              AUTO_EXPANSION_MODE_DEFAULT = 1
          enable_fuzzy_extraction (bool):
            Enables fuzzy entity extraction during
            classification.
          redact (bool):
            Indicates whether parameters of the entity
            type should be redacted in log. If redaction is
            enabled, page parameters and intent parameters
            referring to the entity type will be replaced by
            parameter name during logging.
          overwrite (bool)
            Overwrite the new proto_obj if proto_obj already
            contains an EntityType.

        Returns:
          An EntityType object stored in proto_obj
        """
        if not isinstance(display_name, str):
            raise ValueError(
                "display_name should be a string."
            )
        if not (isinstance(kind, int) and 1 <= kind <= 3):
            raise ValueError(
                "kind should be an int between 0 and 3."
            )
        if self.proto_obj and not overwrite:
            raise UserWarning(
                "proto_obj already contains an EntityType."
                " If you wish to overwrite it, pass overwrite as True."
            )
        if overwrite or not self.proto_obj:
            self.proto_obj = EntityType(
                display_name=display_name,
                kind=kind,
                auto_expansion_mode=auto_expansion_mode,
                enable_fuzzy_extraction=enable_fuzzy_extraction,
                redact=redact
            )

        return self.proto_obj


    def add_excluded_phrase(
        self, phrase: Union[str, List[str]]
    ) -> EntityType:
        """Add one or more phrases to be excluded in the EntityType.

        Args:
          phrase (str | List[str]):
            A single phrase as a string
            or multiple phrases as a list of strings

        Returns:
          An EntityType object stored in proto_obj
        """
        self._check_proto_obj_attr_exist()

        if isinstance(phrase, str):
            self.proto_obj.excluded_phrases.append(
                EntityType.ExcludedPhrase(value=phrase)
            )
        elif isinstance(phrase, list):
            if not all((isinstance(p, str) for p in phrase)):
                raise ValueError(
                    "Only strings are allowed in the phrase list."
                )

            self.proto_obj.excluded_phrases.extend([
                EntityType.ExcludedPhrase(value=value)
                for value in phrase])

        else:
            raise ValueError(
                "phrase should be either a string or a list of strings."
            )

        return self.proto_obj


    def remove_excluded_phrase(
        self, phrase: Union[str, List[str]]
    ) -> EntityType:
        """Remove one or more phrases from excluded phrases in the EntityType.

        Args:
          phrase (str | List[str]):
            A single phrase as a string
            or multiple phrases as a list of strings

        Returns:
          An EntityType object stored in proto_obj
        """
        self._check_proto_obj_attr_exist()

        excl_phrases = {
            excl_phrase.value
            for excl_phrase in self.proto_obj.excluded_phrases
        }
        if isinstance(phrase, str):
            excl_phrases -= {phrase}
        elif isinstance(phrase, list):
            if not all((isinstance(p, str) for p in phrase)):
                raise ValueError(
                    "Only strings are allowed in the phrase list."
                )
            excl_phrases -= set(phrase)
        else:
            raise ValueError(
                "phrase should be either a string or a list of strings."
            )
        # Remove all excluded phrases and add new ones from the excl_phrases set
        self.proto_obj.excluded_phrases.clear()
        self.proto_obj.excluded_phrases.extend([
            EntityType.ExcludedPhrase(value=excl_phrase)
            for excl_phrase in excl_phrases
        ])

        return self.proto_obj


    def add_entity(
        self, value: str, synonyms: List[str] = None
    ) -> EntityType:
        """Add an entity to the EntityType stored in proto_obj

        Args:
          value (str):
            Required. The primary value associated with this entity
            entry. For example, if the entity type is *vegetable*, the
            value could be *scallions*.
            For ``KIND_MAP`` entity types:
            -  A canonical value to be used in place of synonyms.
          synonyms (List[str]):
            Required only for ``KIND_MAP``.
            A collection of value synonyms. For example, if
            the entity type is *vegetable*, and ``value`` is
            *scallions*, a synonym could be *green onions*.

        Returns:
          An EntityType object stored in proto_obj
        """
        self._check_proto_obj_attr_exist()

        if not isinstance(value, str):
            raise ValueError(
                "value should be string."
            )

        if self.proto_obj.kind.name in ["KIND_LIST", "KIND_REGEXP"]:
            self.proto_obj.entities.append(
                EntityType.Entity(
                    value=value, synonyms=[value]
                )
            )
        elif self.proto_obj.kind.name == "KIND_MAP":
            if not (
                isinstance(synonyms, list) and
                all((isinstance(s, str) for s in synonyms))
            ):
                raise ValueError(
                    "synonyms should be a list of strings."
                )

            self.proto_obj.entities.append(
                EntityType.Entity(
                    value=value, synonyms=synonyms
                )
            )
        else:
            raise ValueError(
                "Entity type's kind is not correct and should be specified."
            )

        return self.proto_obj


    def remove_entity(
        self, value: str, synonyms: List[str] = None
    ) -> EntityType:
        """Remove an entity or synonym from the EntityType stored in proto_obj.
        For `KIND_MAP` if synonym was not provided, the whole entity
        will be removed, otherwise the synonyms associated with
        the value will be removed.

        Args:
          value (str):
            Required. The primary value associated with this entity
            entry to remove. For example, if the entity type is *vegetable*,
            the value could be *scallions*.
            For `KIND_MAP` entity types:
            -  A canonical value to be used in place of synonyms.
          synonyms (List[str]):
            Required only for `KIND_MAP`.
            A collection of value synonyms to be removed. For example, if
            the entity type is *vegetable*, and `value` is
            *scallions*, a synonym could be *green onions*.

        Returns:
          An EntityType object stored in proto_obj
        """
        self._check_proto_obj_attr_exist()

        if not isinstance(value, str):
            raise ValueError(
                "value should be string."
            )

        if not synonyms:
            for idx, entity in enumerate(self.proto_obj.entities):
                if entity.value == value:
                    del self.proto_obj.entities[idx]
                    break
        elif self.proto_obj.kind.name == "KIND_MAP":
            if not (
                isinstance(synonyms, list) and
                all((isinstance(s, str) for s in synonyms))
            ):
                raise ValueError(
                    "synonyms should be a list of strings."
                )

            for entity in self.proto_obj.entities:
                if entity.value == value:
                    new_synonyms = list(set(entity.synonyms) - set(synonyms))
                    entity.synonyms.clear()
                    entity.synonyms.extend(new_synonyms)
                    break

        return self.proto_obj



"""Pytest config file"""

# Copyright 2023 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import pytest

def pytest_addoption(parser):
    """Method to add option for creds in tests."""
    parser.addoption("--creds", action="store")
    parser.addoption("--project_id", action="store")
    parser.addoption("--gcs_bucket", action="store")
    parser.addoption("--agent_id", action="store")


@pytest.fixture(scope="session")
def creds(request):
    """Fixture to share creds across the test class"""
    return request.config.getoption("--creds")

@pytest.fixture(scope="session")
def project_id(request):
    """Fixture to share project across the test class"""
    return request.config.getoption("--project_id")

@pytest.fixture(scope="session")
def gcs_bucket(request):
    """Fixture to share gcs_bucket across the test class"""
    return request.config.getoption("--gcs_bucket")

@pytest.fixture(scope="session")
def agent_id(request):
    return request.config.getoption("agent_id")



"""A set of builder methods to create CX proto resource objects"""

# Copyright 2023 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging

from google.cloud.dialogflowcx_v3beta1.types import Fulfillment
from google.cloud.dialogflowcx_v3beta1.types import TransitionRoute
from google.cloud.dialogflowcx_v3beta1.types import EventHandler
from dfcx_scrapi.builders.builders_common import BuildersCommon
from dfcx_scrapi.builders.fulfillments import FulfillmentBuilder

# logging config
logging.basicConfig(
    level=logging.INFO,
    format="%(asctime)s %(levelname)-8s %(message)s",
    datefmt="%Y-%m-%d %H:%M:%S",
)


class TransitionRouteBuilder(BuildersCommon):
    """Base Class for CX TransitionRoute builder."""

    _proto_type = TransitionRoute
    _proto_type_str = "TransitionRoute"


    def __str__(self) -> str:
        """String representation of the proto_obj."""
        self._check_proto_obj_attr_exist()

        target_str = self._show_target()
        transition_criteria = self._show_transition_criteria()
        fulfillment_str = self._show_fulfillment()

        return (
            f"{target_str}"
            f"\n{transition_criteria}"
            f"\nFulfillment:\n\n{fulfillment_str}"
        )


    def _show_transition_criteria(self) -> str:
        """String representation for the transition criteria of proto_obj."""
        self._check_proto_obj_attr_exist()

        intent_str, cond_str = "Not Specified", "Not Specified"
        if self.proto_obj.intent:
            intent_str = self.proto_obj.intent
        if self.proto_obj.condition:
            cond_str = self.proto_obj.condition
        return (
            "Transition criteria:"
            f"\n\tIntent: {intent_str}"
            f"\n\tCondition: {cond_str}"
        )


    def _show_target(self) -> str:
        """String representation for the target of proto_obj."""
        self._check_proto_obj_attr_exist()

        if self.proto_obj.target_page:
            target_type = "Page"
            target_id = self.proto_obj.target_page
        elif self.proto_obj.target_flow:
            target_type = "Flow"
            target_id = self.proto_obj.target_flow
        else:
            target_type = "Not Specified"
            target_id = "None"
        return f"Target: {target_type}\nTarget ID: {target_id}"


    def _show_fulfillment(self) -> str:
        """String representation for the fulfillment of proto_obj."""
        self._check_proto_obj_attr_exist()

        fulfillment_str = ""
        if self.proto_obj.trigger_fulfillment:
            fulfillment_str = str(
                FulfillmentBuilder(self.proto_obj.trigger_fulfillment)
            )

        return fulfillment_str


    def show_transition_route(self, mode: str = "whole"):
        """Show the proto_obj information.
        Args:
          mode (str):
            Specifies what part of the TransitionRoute to show.
            Options:
              ['target', 'fulfillment',
               'transition criteria' or 'conditions', 'whole']
        """
        self._check_proto_obj_attr_exist()

        if mode == "target":
            print(self._show_target())
        elif mode in ["transition criteria", "conditions"]:
            print(self._show_transition_criteria())
        elif mode == "fulfillment":
            print(self._show_fulfillment())
        elif mode == "whole":
            print(self)
        else:
            raise ValueError(
                "mode should be in"
                " ['target', 'fulfillment',"
                " 'transition criteria' or 'conditions', 'whole']"
            )


    def create_new_proto_obj(
        self,
        intent: str = None,
        condition: str = None,
        trigger_fulfillment: Fulfillment = None,
        target_page: str = None,
        target_flow: str = None,
        overwrite: bool = False
    ) -> TransitionRoute:
        """Create a new TransitionRoute.

        Args:
          intent (str):
            Indicates that the transition can only happen when the given
            intent is matched.
            Format:
            ``projects/<Project ID>/locations/<Location ID>/
              agents/<Agent ID>/intents/<Intent ID>``.
            At least one of ``intent`` or ``condition`` must be specified.
            When both ``intent`` and ``condition`` are specified,
            the transition can only happen when both are fulfilled.
          condition (str):
            The condition to evaluate.
            See the conditions reference:
            https://cloud.google.com/dialogflow/cx/docs/reference/condition
            At least one of ``intent`` or ``condition`` must be specified.
            When both ``intent`` and ``condition`` are specified,
            the transition can only happen when both are fulfilled.
          trigger_fulfillment (Fulfillment):
            The fulfillment to call when the condition is satisfied.
            When ``trigger_fulfillment`` and ``target`` are defined,
            ``trigger_fulfillment`` is executed first.
          target_page (str):
            The target page to transition to. Format:
            ``projects/<Project ID>/locations/<Location ID>/
              agents/<Agent ID>/flows/<Flow ID>/pages/<Page ID>``.
            At most one of ``target_page`` and ``target_flow``
            can be specified at the same time.
          target_flow (str):
            The target flow to transition to. Format:
            ``projects/<Project ID>/locations/<Location ID>/
              agents/<Agent ID>/flows/<Flow ID>``.
            At most one of ``target_page`` and ``target_flow``
            can be specified at the same time.
          overwrite (bool)
            Overwrite the new proto_obj if proto_obj already
            contains a TransitionRoute.

        Returns:
          A TransitionRoute object stored in proto_obj.
        """
        # Types error checking
        if ((intent and not isinstance(intent, str)) or
            (condition and not isinstance(condition, str)) or
            (target_page and not isinstance(target_page, str)) or
            (target_flow and not isinstance(target_flow, str))):
            raise ValueError(
                "intent, condition, target_page, and target_flow"
                " if existed should be a string."
            )
        if (trigger_fulfillment and
            not isinstance(trigger_fulfillment, Fulfillment)):
            raise ValueError(
                "The type of trigger_fulfillment should be a Fulfillment."
            )
        # Minimum requirement error checking
        if not(intent or condition):
            raise RuntimeError(
                "At least one of `intent` or `condition` must be specified."
            )
        # Oneof error checking
        if target_page and target_flow:
            raise RuntimeError(
                "At most one of `target_page` and `target_flow`"
                " can be specified at the same time."
            )
        # `overwrite` parameter error checking
        if self.proto_obj and not overwrite:
            raise UserWarning(
                "proto_obj already contains a TransitionRoute."
                " If you wish to overwrite it, pass overwrite as True."
            )

        # Create the TransitionRoute
        if overwrite or not self.proto_obj:
            # Create empty Fulfillment in case user didn't pass any
            if not trigger_fulfillment:
                trigger_fulfillment = Fulfillment()

            self.proto_obj = TransitionRoute(
                intent=intent,
                condition=condition,
                trigger_fulfillment=trigger_fulfillment,
                target_page=target_page,
                target_flow=target_flow
            )

        return self.proto_obj




class EventHandlerBuilder(BuildersCommon):
    """Base Class for CX EventHandler builder."""

    _proto_type = EventHandler
    _proto_type_str = "EventHandler"


    def __str__(self) -> str:
        """String representation of the proto_obj."""
        self._check_proto_obj_attr_exist()

        event_and_target_str = self._show_event_and_target()
        fulfillment_str = self._show_fulfillment()

        return (
            f"{event_and_target_str}"
            f"\nFulfillment:\n\n{fulfillment_str}"
        )


    def _show_event_and_target(self) -> str:
        """String representation for the target of proto_obj."""
        self._check_proto_obj_attr_exist()

        # Event str
        event_str = f"Event: {self.proto_obj.event}"
        # Target str
        if self.proto_obj.target_page:
            target_type = "Page"
            target_id = self.proto_obj.target_page
        elif self.proto_obj.target_flow:
            target_type = "Flow"
            target_id = self.proto_obj.target_flow
        else:
            target_type = "Not Specified"
            target_id = "None"
        return (
            f"{event_str}"
            f"\nTarget: {target_type}"
            f"\nTarget ID: {target_id}"
        )


    def _show_fulfillment(self) -> str:
        """String representation for the fulfillment of proto_obj."""
        self._check_proto_obj_attr_exist()

        fulfillment_str = ""
        if self.proto_obj.trigger_fulfillment:
            fulfillment_str = str(
                FulfillmentBuilder(self.proto_obj.trigger_fulfillment)
            )

        return fulfillment_str


    def show_event_handler(self, mode: str = "whole"):
        """Show the proto_obj information.
        Args:
          mode (str):
            Specifies what part of the EventHandler to show.
            Options:
              ['basic' or 'target' or 'event', 'fulfillment', 'whole']
        """
        self._check_proto_obj_attr_exist()

        if mode in ["basic", "target", "event"]:
            print(self._show_event_and_target())
        elif mode == "fulfillment":
            print(self._show_fulfillment())
        elif mode == "whole":
            print(self)
        else:
            raise ValueError(
                "mode should be in"
                " ['basic' or 'target' or 'event', 'fulfillment', 'whole']"
            )


    def create_new_proto_obj(
        self,
        event: str,
        trigger_fulfillment: Fulfillment = None,
        target_page: str = None,
        target_flow: str = None,
        overwrite: bool = False
    ) -> EventHandler:
        """Create a new EventHandler.

        Args:
          event (str):
            Required. The name of the event to handle.
          trigger_fulfillment (Fulfillment):
            The fulfillment to call when the event occurs.
            Handling webhook errors with a fulfillment enabled with webhook
            could cause infinite loop. It is invalid to specify
            such fulfillment for a handler handling webhooks.
          target_page (str):
            The target page to transition to. Format:
            ``projects/<Project ID>/locations/<Location ID>/
              agents/<Agent ID>/flows/<Flow ID>/pages/<Page ID>``.
            At most one of ``target_page`` and ``target_flow``
            can be specified at the same time.
          target_flow (str):
            The target flow to transition to. Format:
            ``projects/<Project ID>/locations/<Location ID>/
              agents/<Agent ID>/flows/<Flow ID>``.
            At most one of ``target_page`` and ``target_flow``
            can be specified at the same time.
          overwrite (bool)
            Overwrite the new proto_obj if proto_obj already
            contains a EventHandler.

        Returns:
          An EventHandler object stored in proto_obj.
        """
        # Types error checking
        if event and not isinstance(event, str):
            raise ValueError("event should be a string.")
        if (trigger_fulfillment and
            not isinstance(trigger_fulfillment, Fulfillment)):
            raise ValueError(
                "The type of trigger_fulfillment should be a Fulfillment."
            )
        # Oneof error checking
        if target_page and target_flow:
            raise RuntimeError(
                "At most one of `target_page` and `target_flow`"
                " can be specified at the same time."
            )
        # `overwrite` parameter error checking
        if self.proto_obj and not overwrite:
            raise UserWarning(
                "proto_obj already contains an EventHandler."
                " If you wish to overwrite it, pass overwrite as True."
            )
        # Create the EventHandler
        if overwrite or not self.proto_obj:
            # Create empty Fulfillment in case user didn't pass any
            if not trigger_fulfillment:
                trigger_fulfillment = Fulfillment()

            self.proto_obj = EventHandler(
                event=event,
                trigger_fulfillment=trigger_fulfillment,
                target_page=target_page,
                target_flow=target_flow
            )

        return self.proto_obj



"""Common methods and helper functions used throughout library."""

# Copyright 2023 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging
import re
from dfcx_scrapi.agent_extract import types

# logging config
logging.basicConfig(
    level=logging.INFO,
    format="%(message)s",
)

class Common:
    """Common methods and helper functions used throughout library."""

    @staticmethod
    def parse_filepath(in_path: str, resource_type: str) -> str:
        """Parse file path to provide quick reference for resource."""

        regex_map = {
            "flow": r".*\/flows\/([^\/]*)",
            "page": r".*\/pages\/([^\/]*)\.",
            "entity_type": r".*\/entityTypes\/([^\/]*)",
            "intent": r".*\/intents\/([^\/]*)",
            "route_group": r".*\/transitionRouteGroups\/([^\/]*)",
            "webhook": r".*\/webhooks\/([^\/]*)\."
        }
        resource_name = re.match(regex_map[resource_type], in_path).groups()[0]

        return resource_name

    @staticmethod
    def clean_display_name(display_name: str):
        """Replace cspecial haracters from map for the given display name."""
        patterns = {
            "%22": '"',
            "%23": "#",
            "%24": "$",
            "%26": "&",
            "%27": "'",
            "%28": "(",
            "%29": ")",
            "%2b": "+",
            "%2c": ",",
            "%2f": "/",
            "%3a": ":",
            "%3c": "<",
            "%3d": "=",
            "%3e": ">",
            "%3f": "?",
            "%5b": "[",
            "%5d": "]",
            "%e2%80%9c": "“",
            "%e2%80%9d": "”",
        }

        for key, value in patterns.items():
            if key in display_name:
                display_name = display_name.replace(key, value)

        return display_name

    @staticmethod
    def check_lang_code(lang_code: str, stats: types.AgentData):
        """Check to see if file lang_code matches user input lang_code."""
        return stats.lang_code == lang_code



"""CX Transition Route Group Resource functions."""

# Copyright 2023 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging
import time
from typing import Dict
import pandas as pd
from google.cloud.dialogflowcx_v3beta1 import services
from google.cloud.dialogflowcx_v3beta1 import types
from google.protobuf import field_mask_pb2

from dfcx_scrapi.core import flows
from dfcx_scrapi.core import intents
from dfcx_scrapi.core import pages
from dfcx_scrapi.core import webhooks
from dfcx_scrapi.core import scrapi_base


# logging config
logging.basicConfig(
    level=logging.INFO,
    format="%(asctime)s %(levelname)-8s %(message)s",
    datefmt="%Y-%m-%d %H:%M:%S",
)


class TransitionRouteGroups(scrapi_base.ScrapiBase):
    """Core Class for CX Transition Route Group functions."""

    def __init__(
        self,
        creds_path: str = None,
        creds_dict: Dict = None,
        creds=None,
        scope=False,
        route_group_id: str = None,
        flow_id: str = None,
        agent_id: str = None,
    ):
        super().__init__(
            creds_path=creds_path,
            creds_dict=creds_dict,
            creds=creds,
            scope=scope,
        )

        self.flows = flows.Flows(creds=self.creds)
        self.intents = intents.Intents(creds=self.creds)
        self.pages = pages.Pages(creds=self.creds)
        self.webhooks = webhooks.Webhooks(creds=self.creds)

        if route_group_id:
            self.route_group_id = route_group_id
            self.client_options = self._set_region(route_group_id)

        if flow_id:
            self.flow_id = flow_id

        if agent_id:
            self.agent_id = agent_id

    def _rg_temp_dict_update(self, temp_dict, element):
        """Modify the temp dict and return to dataframe function."""
        element_dict = self.cx_object_to_dict(element)
        key = list(element_dict.keys())[0]

        if key == "payload":
            temp_dict.update({"custom_payload": element_dict[key]})
        elif key == "liveAgentHandoff":
            temp_dict.update(
                {"live_agent_handoff": element_dict[key]["metadata"]}
            )
        elif key == "conversationSuccess":
            temp_dict.update(
                {"conversation_success": element_dict[key]["metadata"]}
            )
        elif key == "playAudio":
            temp_dict.update({"play_audio": element_dict[key]["audioUri"]})
        elif key == "outputAudioText":
            temp_dict.update({"output_audio_text": element_dict[key]["text"]})
        elif key == "text":
            if len(element_dict[key]["text"]) == 1:
                temp_dict.update(
                    {"fulfillment_message": element_dict[key]["text"][0]}
                )
            else:
                temp_dict.update(
                    {"fulfillment_message": element_dict[key]["text"]}
                )
        else:
            temp_dict.update({key: element_dict[key]})

        return temp_dict

    def get_route_groups_map(self, flow_id: str = None, reverse=False):
        """Exports Agent Route Group UUIDs and Names into a user friendly dict.

        Args:
          flow_id: the formatted CX Agent Flow ID to use
          reverse: (Optional) Boolean flag to swap key:value -> value:key

        Returns:
          Dictionary containing Route Group UUIDs as keys and display names
          as values. If Optional reverse=True, the output will return
          route group name:ID mapping instead of ID:route group name
        """
        if not flow_id:
            flow_id = self.flow_id

        if reverse:
            pages_dict = {
                page.display_name: page.name
                for page in self.list_transition_route_groups(flow_id)
            }

        else:
            pages_dict = {
                page.name: page.display_name
                for page in self.list_transition_route_groups(flow_id)
            }

        return pages_dict

    @scrapi_base.api_call_counter_decorator
    def list_transition_route_groups(self, flow_id: str = None):
        """Exports List of all Route Groups in the specified CX Flow ID.

        Args:
          flow_id: The formatted CX Flow ID to list the route groups from

        Returns:
          List of Route Group objects
        """
        if not flow_id:
            flow_id = self.flow_id

        request = (
            types.transition_route_group.ListTransitionRouteGroupsRequest()
        )
        request.parent = flow_id

        client_options = self._set_region(flow_id)
        client = services.transition_route_groups.TransitionRouteGroupsClient(
            credentials=self.creds, client_options=client_options
        )
        response = client.list_transition_route_groups(request)

        cx_route_groups = []
        for page in response.pages:
            for cx_route_group in page.transition_route_groups:
                cx_route_groups.append(cx_route_group)

        return cx_route_groups

    @scrapi_base.api_call_counter_decorator
    def get_transition_route_group(self, route_group_id):
        """Get a single Transition Route Group object.

        Args:
          route_group_id: the formatted CX Route Group ID to retrieve.

        Returns:
          A single Route Group object
        """
        request = types.transition_route_group.GetTransitionRouteGroupRequest()
        request.name = route_group_id
        client_options = self._set_region(route_group_id)
        client = services.transition_route_groups.TransitionRouteGroupsClient(
            credentials=self.creds, client_options=client_options
        )
        response = client.get_transition_route_group(request)

        return response

    @scrapi_base.api_call_counter_decorator
    def create_transition_route_group(
        self,
        flow_id: str = None,
        obj: types.TransitionRouteGroup = None,
        **kwargs,
    ):
        """Create a single Transition Route Group resource.

        Args:
          flow_id: the formatted CX Flow ID to create the route group in
          obj: (Optional) the Transition Route Group object of type
            types.TransitionRouteGroup that you want the new route group
            to be built from.

        Returns:
          A copy of the successfully created Route Group object
        """
        if not flow_id:
            flow_id = self.flow_id

        if obj:
            trg = obj
            trg.name = ""
        else:
            trg = types.transition_route_group.TransitionRouteGroup()

        for key, value in kwargs.items():
            setattr(trg, key, value)

        client_options = self._set_region(flow_id)
        client = services.transition_route_groups.TransitionRouteGroupsClient(
            credentials=self.creds, client_options=client_options
        )
        response = client.create_transition_route_group(
            parent=flow_id, transition_route_group=trg
        )

        return response

    @scrapi_base.api_call_counter_decorator
    def update_transition_route_group(
        self,
        route_group_id: str = None,
        obj: types.TransitionRouteGroup = None,
        language_code: str = None,
        **kwargs,
    ):
        """Update a single Route Group resource.

        Args:
          route_group_id: the formatted CX Route Group ID to update.
          obj: (Optional) the Transition Route Group object of type
            types.TransitionRouteGroup that you want to update.
          language_code: (Optional) the language in which the agent should
            update the TransitionRouteGroup

        Returns:
          A copy of the successfully updated Route Group object
        """
        if obj:
            route_group = obj
            route_group.name = route_group_id
        else:
            route_group = self.get_transition_route_group(route_group_id)

        for key, value in kwargs.items():
            setattr(route_group, key, value)
        paths = kwargs.keys()
        mask = field_mask_pb2.FieldMask(paths=paths)

        client_options = self._set_region(route_group_id)
        client = services.transition_route_groups.TransitionRouteGroupsClient(
            credentials=self.creds, client_options=client_options
        )

        request = (
            types.transition_route_group.UpdateTransitionRouteGroupRequest()
        )  # pylint: disable=C0301

        request.transition_route_group = route_group
        request.update_mask = mask

        if language_code:
            request.language_code = language_code

        response = client.update_transition_route_group(request)

        return response

    def route_groups_to_dataframe(
        self, agent_id: str = None, rate_limit: float = 0.5
    ):
        """Extracts the Transition Route Groups from a given Agent and
         returns key information about the Route Groups in a Pandas Dataframe

        DFCX Route Groups exist as an Agent level resource, however they are
        categorized by the Flow they are associated with. This method will
        extract all Flows for the given agent, then use the Flow IDs to
        extract all Route Groups per Flow. Once all Route Groups have been
        extracted, the method will convert the DFCX object to a Pandas
        Dataframe and return this to the user.

        Args:
          agent_id: the Agent ID string in the following format:
            projects/<project_id>/locations/<location_id>/agents/<agent_id>
          rate_limit: Time in seconds to wait between each API call. Use this
            to control hitting Quota limits on your project.

        Returns:
          a Pandas Dataframe with columns: flow, route_group_name, target_page,
          intent, condition, webhook, webhook_tag, custom_payload,
          live_agent_handoff, conversation_success, play_audio,
          output_audio_text, fulfillment_message
        """
        if not agent_id:
            agent_id = self.agent_id

        flows_map = self.flows.get_flows_map(agent_id)
        intents_map = self.intents.get_intents_map(agent_id)
        webhooks_map = self.webhooks.get_webhooks_map(agent_id)
        all_pages_map = {}
        all_rgs = []

        for flow in flows_map:
            all_pages_map.update(self.pages.get_pages_map(flow))
            all_rgs.extend(self.list_transition_route_groups(flow))
            time.sleep(rate_limit)

        rows_list = []
        for route_group in all_rgs:
            flow = "/".join(route_group.name.split("/")[0:8])
            for route in route_group.transition_routes:
                temp_dict = {}

                temp_dict.update({"flow": flows_map[flow]})
                temp_dict.update({"route_group_name": route_group.display_name})

                if route.target_page:
                    t_p = all_pages_map.get(route.target_page)
                    if not t_p:
                        t_p = str(route.target_page).rsplit("/", maxsplit=1)[-1]

                    temp_dict.update({"target_page": t_p})

                if route.intent:
                    temp_dict.update({"intent": intents_map[route.intent]})

                if route.condition:
                    temp_dict.update({"condition": route.condition})

                if route.trigger_fulfillment.webhook:
                    temp_dict.update(
                        {
                            "webhook": webhooks_map[
                                route.trigger_fulfillment.webhook
                            ]
                        }
                    )

                    temp_dict.update(
                        {"webhook_tag": route.trigger_fulfillment.tag}
                    )

                if route.trigger_fulfillment.messages:
                    for element in route.trigger_fulfillment.messages:
                        temp_dict = self._rg_temp_dict_update(
                            temp_dict, element
                        )

                rows_list.append(temp_dict)

        final_dataframe = pd.DataFrame(rows_list)

        return final_dataframe



"""Flow extract methods and functions."""

# Copyright 2023 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import json
import os

from typing import List

from dfcx_scrapi.agent_extract import graph
from dfcx_scrapi.agent_extract import common
from dfcx_scrapi.agent_extract import types
from dfcx_scrapi.agent_extract import pages
from dfcx_scrapi.agent_extract import routes
from dfcx_scrapi.agent_extract import route_groups


class Flows:
    """Flow processing methods and functions."""

    def __init__(self):
        self.common = common.Common()
        self.pages = pages.Pages()
        self.rgs = route_groups.RouteGroups()
        self.routes = routes.Fulfillments()
        self.special_pages = [
            "End Session",
            "End Flow",
            "Start Page",
            "Current Page",
            "Previous Page",
        ]

    @staticmethod
    def build_flow_path_list(agent_local_path: str):
        """Builds a list of dirs, each representing a Flow directory.

        Ex: /path/to/agent/flows/<flow_dir>

        This dir path can then be used to find the next level of information
        in the directory by appending the appropriate next dir structures like:
        - <flow_name>.json, for the Flow object
        - /transitionRouteGroups, for the Route Groups dir
        - /pages, for the Pages dir
        """
        root_dir = agent_local_path + "/flows"

        flow_paths = []

        for flow_dir in os.listdir(root_dir):
            flow_dir_path = f"{root_dir}/{flow_dir}"
            flow_paths.append(flow_dir_path)

        return flow_paths

    @staticmethod
    def remove_flow_pages_from_set(input_set: set) -> set:
        """Remove any transitions tagged with FLOW.

        Some route transitions go to Flow instead of Page. For these
        transitions, we tag them with `FLOW` for easier identification later.
        However, when reporting on Graph inconsistencies like Dangling or
        Unreachable pages, we want to remove these from any result sets as they
        are not relevant.
        """
        filtered_set = set()

        for page in input_set:
            if "FLOW" not in page:
                filtered_set.add(page)

        return filtered_set

    def find_unreachable_pages(self, flow: types.Flow):
        """Find Unreachable Pages in the graph.

        An Unreachable Page is defined as:
          - A Page which has no incoming edge when traversed from Start Page.
            That is, it is unreachable in the graph by any practical means.
          - A Page which is connected to a root unreachable page. That is, a
            page that could have both incoming or outgoing routes, but due to
            its connectedness to the root orphan page, is unreachable in the
            graph.

        Here we will compute the symmetric difference of 2 sets:
          - Active Pages (i.e. Pages that were reachable in the graph)
          - Used Pages (i.e. Pages that were used by some Route)

        If an Unreachable Page has children that it routes to, those children
        will appear in Used Pages, although they will ultimately be
        unreachable. It's possible for an Unreachable Page to route back to an
        Active Page in the graph. For these instances, we don't want to count
        those pages as unreachable, because they are reachable via other
        sections of the graph.
        """
        filtered_set = flow.active_pages.symmetric_difference(
            flow.graph.used_nodes
        )
        filtered_set = self.remove_flow_pages_from_set(filtered_set)
        flow.unreachable_pages.update(filtered_set)

        return flow

    def find_unused_pages(self, flow: types.Flow):
        """Find Unused Pages in the graph.

        An Unused Page is defined as:
          - A Page which has no incoming or outgoing edge AND
          - A Page which exists in the Agent design time, but which is not
            present anywhere in the graph, either visible or non-visible.

        Here we will compute the difference of 2 sets:
          - All Pages (i.e. Pages that exist in the Agent Design Time)
          - Used Pages (i.e. Pages that were used by some Route)

        The resulting set will consist of 2 types of Pages:
          - Truly Unused Pages
          - Unreachable Root Pages

        Unreachable Root Pages end up in the results due to the fact that no
        other Active Page is pointing to them. We remove these from the
        resulting set before presenting the Truly Unused Pages.
        """

        # Discard special pages as they are non-relevant for final outcome
        for page in self.special_pages:
            flow.all_pages.discard(page)

        prelim_unused = flow.all_pages.difference(flow.graph.used_nodes)

        # Filter out Unreachable Root Pages
        filtered_set = set()

        for page in prelim_unused:
            if page not in flow.graph.edges:
                filtered_set.add(page)
            else:
                flow.unreachable_pages.add(page)

        flow.unused_pages = filtered_set

        return flow

    def recurse_edges(
        self, edges: List, page: types.Page, dangling: set, visited: set
    ):
        """Recursive method searching graph edges for Active / Dangling Pages.

        A byproduct of searching for Dangling Pages in the graph is that we can
        produce a set of Active Pages in the graph. These are pages that are
        reachable when traversing from the Start Page. These can then be used
        to determine Unreachable Pages in another method.
        """
        # For Flow Start Pages, we prepend the Flow name for later
        # identification. For this section, we'll need to strip it off to
        # compare with the other sets.
        if page in edges:
            for inner_page in edges[page]:
                if inner_page not in visited:
                    visited.add(inner_page)
                    dangling, visited = self.recurse_edges(
                        edges, inner_page, dangling, visited
                    )

        else:
            dangling.add(page)

        return dangling, visited

    def find_dangling_pages(self, flow: types.Flow):
        """Find Dangling Pages in the graph.

        Dangling Page is defined as:
          - Any page that exists in the graph that has no outgoing edge
        Active Page is defined as:
          - Any page that is reachable via an active route in the graph and can
            be traced back to the Start Page.

        These pages can result in a conversational "dead end" which is
        potentially unrecoverable.
        A byproduct of searching for the dangling pages is locating all of the
        "active" pages. These are the pages that are "visited" as we traverse
        the graph. We'll also return Active Pages in this method since they
        will be used for downstream tasks.
        """

        flow.dangling_pages, flow.active_pages = self.recurse_edges(
            flow.graph.edges,
            f"{flow.display_name}: Start Page",
            flow.dangling_pages,
            flow.active_pages,
        )

        # Clean up Special Pages
        for page in self.special_pages:
            flow.dangling_pages.discard(page)

        flow.dangling_pages = self.remove_flow_pages_from_set(
            flow.dangling_pages
        )

        return flow

    def process_start_page(self, flow: types.Flow, stats: types.AgentData):
        """Process a single Flow Path file."""
        with open(flow.start_page_file, "r", encoding="UTF-8") as flow_file:
            page = types.Page(flow=flow)
            page.display_name = f"{flow.display_name}: Start Page"

            # We keep track of an instance specific Flow graph for the current
            # Flow, and then a main Graph for the entire agent.
            flow.graph.add_node(page.display_name)
            stats.graph.add_node(page.display_name)

            page.data = json.load(flow_file)
            page.events = page.data.get("eventHandlers", None)
            page.routes = page.data.get("transitionRoutes", None)
            page.route_groups = page.data.get("transitionRouteGroups", None)
            stats.flows.append(page.data)

            flow.resource_id = page.data.get("name", None)

            # Order of processing is important
            stats = self.routes.process_routes(page, stats)
            stats = self.routes.process_events(page, stats)

            if page.route_groups:
                page, stats = self.routes.set_route_group_targets(page, stats)

            flow_file.close()

        full_flow_id = f"{stats.agent_id}/flows/{flow.resource_id}"
        stats.flows_map[flow.display_name] = full_flow_id
        stats.flow_page_map[flow.display_name] = {
            "id": full_flow_id,
            "pages": {}
            }

        return stats

    def process_flow(self, flow: types.Flow, stats: types.AgentData):
        """Process a Single Flow dir and all subdirectories."""
        flow.file_name = self.common.parse_filepath(flow.dir_path, "flow")
        flow.display_name = self.common.clean_display_name(flow.file_name)

        flow.start_page_file = f"{flow.dir_path}/{flow.file_name}.json"

        stats.pages[flow.display_name] = []
        stats.active_intents[flow.display_name] = []
        stats = self.process_start_page(flow, stats)
        stats = self.pages.process_pages_directory(flow, stats)
        stats = self.rgs.process_route_groups_directory(flow, stats)

        # Order of Find Operations is important here!
        flow = self.find_unused_pages(flow)
        flow = self.find_dangling_pages(flow)
        flow = self.find_unreachable_pages(flow)

        stats.active_pages[flow.display_name] = flow.active_pages
        stats.unused_pages[flow.display_name] = flow.unused_pages
        stats.unreachable_pages[flow.display_name] = flow.unreachable_pages

        return stats

    def process_flows_directory(
            self, agent_local_path: str, stats: types.AgentData):
        """Process the top level Flows dir in the JSON Package structure.

        The following files/dirs exist under the `flows` dir:
        - Flow object (i.e. Flow START_PAGE)
        - transitionRouteGroups
        - pages

        In Dialogflow CX, the START_PAGE of each Flow is a special kind of Page
        that exists within the Flow object itself. In this method, we will lint
        the Flow object, all files in the transitionRouteGroups dir and all
        files in the pages dir.
        """
        # Create a list of all Flow paths to iter through
        flow_paths = self.build_flow_path_list(agent_local_path)
        stats.total_flows = len(flow_paths)

        for flow_path in flow_paths:
            flow = types.Flow()
            flow.graph = graph.Graph()
            flow.dir_path = flow_path
            stats = self.process_flow(flow, stats)

        return stats



"""Agent Assist Resource functions."""

# Copyright 2023 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging

from dfcx_scrapi.core import scrapi_base
from typing import Dict
from google.cloud.dialogflow_v2beta1 import services
from google.cloud.dialogflow_v2beta1 import types
from google.cloud.dialogflow_v2beta1.services.knowledge_bases import KnowledgeBasesClient
from google.protobuf import field_mask_pb2

# logging config
logging.basicConfig(
    level=logging.INFO,
    format="%(asctime)s %(levelname)-8s %(message)s",
    datefmt="%Y-%m-%d %H:%M:%S",
)


class AgentAssist(scrapi_base.ScrapiBase):
    """Core Class for Agent Assist Resource functions."""

    def __init__(
        self,
        creds_path: str = None,
        creds_dict: Dict = None,
        creds=None,
        scope=False,
        conversation_profile_id: str = None,
    ):
        super().__init__(
            creds_path=creds_path,
            creds_dict=creds_dict,
            creds=creds,
            scope=scope,
        )
        self.conversation_profile_id = None

        if conversation_profile_id:
            self.conversation_profile_id = conversation_profile_id

    def _set_conversation_profile(
        self,
        conversation_profile_id: str,
        project_id: str = None
    ):
        """Updating the conversation profile for the object
        Args:
            conversation_profile_id (reqd):
                Unique identifier for the conversation profile.
            project_id (str):
                Optional Field. Required if the full conv path is not supplied.
        """
        logging.info(f"Setting Conversation Profile: {conversation_profile_id}")

        if len(conversation_profile_id.split("/")) >= 4:
            profile_path = conversation_profile_id
        else:
            logging.info("Inferring Profile Path using the Project ID")
            if not project_id:
                raise ValueError(
                    "Project ID is required if the full conevrsation profile"
                    " path is not provided."
                )
            project_path = f"projects/{project_id}"
            conv_path = f"/conversationProfiles/{conversation_profile_id}"
            profile_path = project_path + conv_path

        logging.info(f"Conversation Profile Path :{profile_path}")
        self.conversation_profile_id = profile_path

    def create_conversation_profile(
        self,
        display_name,
        knowledge_base_id: str,
        project_id,
        suggestion_type: str,
        no_small_talk: bool = True,
        only_end_user: bool = True,
        max_results: int = 3,
        language_code: str = "en-US") -> types.ConversationProfile:
        """ Create & configure a conversation profile as per the given input.

        Args:
            display_name (str):
                Required. Human readable name for this
                profile. Max length 1024 bytes.
            project_id (str):
                Required. The project to create a conversation profile for.
            knowledge_base_id (str):
                The knowledge base id of the knowledge base to be configured
                with the conversation profile.
            suggestion_type (str):
                The agent assist feature to be configured with the
                conversation profile.
                Allowed values are: ARTICLE_SUGGESTION, FAQ & SMART_REPLY.
            no_small_talk (bool):
                Do not trigger if last utterance is small talk.
            only_end_user (bool):
                Only trigger suggestion if participant role of last
                utterance is END_USER.
            max_results (int):
                Maximum number of results to return.
                Currently, if unset, defaults to 10. And the max number is 20.
            language_code (str):
                Language code for the conversation profile. If not
                specified, the language is en-US. Language at
                ConversationProfile should be set for all non en-us
                languages. This should be a
                `BCP-47 <https://www.rfc-editor.org/rfc/bcp/bcp47.txt>`__
                language tag. Example: "en-US".

        Returns:
            Conversation profile to be used for incoming
            Dialogueflow conversations.
            type: google.cloud.dialogflow_v2beta1.types.ConversationProfile
        """

        client = services.conversation_profiles.ConversationProfilesClient(
            credentials = self.creds
            )

        project_path = client.common_project_path(project_id)
        logging.info(f"Project Path : {project_path}")

        conversation_profile = types.ConversationProfile()
        conversation_profile.display_name = display_name
        conversation_profile.language_code = language_code

        hac_config = types.HumanAgentAssistantConfig
        feature_config = hac_config.SuggestionFeatureConfig()

        # Configuring Suggestion Type
        suggestion_feature = types.SuggestionFeature()
        suggestion_feature.type_ = suggestion_type

        feature_config.suggestion_feature = suggestion_feature

        # Configuring the trigger settings
        trigger_settings = hac_config.SuggestionTriggerSettings()
        trigger_settings.no_small_talk = no_small_talk
        trigger_settings.only_end_user = only_end_user

        feature_config.suggestion_trigger_settings = trigger_settings

        # Configuring the query config.
        query_config = hac_config.SuggestionQueryConfig()

        kb_path = KnowledgeBasesClient.knowledge_base_path(
            project_id, knowledge_base_id
            )
        logging.info(f"Knowledge Base Path : {kb_path}")
        query_config.knowledge_base_query_source.knowledge_bases=[kb_path]

        query_config.max_results = max_results
        feature_config.query_config = query_config

        conversation_profile.human_agent_assistant_config.human_agent_suggestion_config.feature_configs = [feature_config] # pylint: disable=C0301

        request = types.conversation_profile.CreateConversationProfileRequest(
                    parent= project_path,
                    conversation_profile=conversation_profile,
                )

        # Make the request
        response = client.create_conversation_profile(request=request)
        logging.info("Conversation Profile created...")
        logging.info(f"Display Name: {response.display_name}")
        logging.info(f"Name: {response.name}")

        return response

    def get_conversation_profile(
        self,
        conversation_profile_id: str = None
    ) -> types.ConversationProfile:
        """Returns the conversation profile object
        Args:
            conversation_profile_id: str
                Optional Field. The profile id for requested object.

        Returns:
            Conversation profile Object.
            type: google.cloud.dialogflow_v2beta1.types.ConversationProfile
        """
        client = services.conversation_profiles.ConversationProfilesClient(
            credentials = self.creds
            )
        if not conversation_profile_id:
            conversation_profile_id = self.conversation_profile_id

        request = types.conversation_profile.GetConversationProfileRequest(
                    name=conversation_profile_id,
                )

        response = client.get_conversation_profile(request=request)

        return response

    def list_conversation_profiles(
        self,
        project_id: str
    ):
        """List all the existing conversation profile in the given project
        Args:
            project_id :
            Required. The project to list the conversation profile for.

        Returns:
            List of tuples [(profile display name, profile name)]
        """
        client = services.conversation_profiles.ConversationProfilesClient(
            credentials = self.creds
            )

        project_path = client.common_project_path(project_id)
        logging.info(f"Project Path : {project_path}")

        response = client.list_conversation_profiles(parent = project_path)

        profile_list = []
        for profile in response:
            profile_list.append((profile.display_name, profile.name))

        return profile_list

    def delete_conversation_profile(
        self,
        conversation_profile_id: str
    ):
        """Delete the given conversation profile
        Args:
            conversation_profile_id:
                The unique identifier of this conversation profile. Format:
                ``projects/<Project ID>/locations/<Location ID>
                  /conversationProfiles/<Conversation Profile ID>``.
        """
        client = services.conversation_profiles.ConversationProfilesClient(
            credentials = self.creds
            )

        logging.info(f"Deleting : {conversation_profile_id}")

        client.delete_conversation_profile(name = conversation_profile_id)

    def update_conversation_profile(
        self,
        conversation_profile_id: str = None,
        obj: types.ConversationProfile = None,
        **kwargs
    ) -> types.ConversationProfile:
        """
        Updates the conversation profile object based on the provided args.
        Args:
            conversation_profile_id: str
                The unique identifier of this conversation profile.
                Format:
                ``projects/<Project ID>/locations/<Location ID>/
                  conversationProfiles/<Conversation Profile ID>``.
            obj: The Conversation Profile object to be updated.

        Returns:
            The updated conversation profile object.
        """
        if obj:
            conversation_profile = obj
            conversation_profile.name = conversation_profile_id

        else:
            if not conversation_profile_id:
                conversation_profile_id = self.conversation_profile_id

            conversation_profile = self.get_conversation_profile(
                conversation_profile_id = conversation_profile_id
                )

        # set conversation profile attributes from kwargs
        for key, value in kwargs.items():
            setattr(conversation_profile, key, value)
            logging.info(f"Updating field {key} to {value}")

        paths = kwargs.keys()
        mask = field_mask_pb2.FieldMask(paths=paths)

        client = services.conversation_profiles.ConversationProfilesClient(
            credentials = self.creds
            )

        request = types.UpdateConversationProfileRequest()

        request.conversation_profile = conversation_profile
        request.update_mask = mask

        response = client.update_conversation_profile(request)

        return response

    def create_conversation(
        self,
        project_id: str,
        conversation_profile_id: str = None
        ) -> types.Conversation:
        """Creates a conversation with given values

        Args:
            project_id (str):
                Required. The project to create a conversation profile for.
            conversation_profile_id (str):
                Optional: If not set before, user can pass a profile id to
                be used for creating a conversation.

        Returns:
            Response for the created conversation.
            type: google.cloud.dialogflow_v2beta1.types.Conversation
        """

        client = services.conversations.ConversationsClient(
            credentials = self.creds
            )

        if conversation_profile_id:
            self._set_conversation_profile(
                conversation_profile_id = conversation_profile_id,
                project_id = project_id
                )

        if not self.conversation_profile_id and not conversation_profile_id:
            raise ValueError(
                "Conversation Profile ID required, as it isn't set for this"
                " instance of the AgentAssist class."
            )

        project_path = client.common_project_path(project_id)

        conversation = {"conversation_profile": self.conversation_profile_id}
        response =  client.create_conversation(
            parent=project_path,
            conversation=conversation
            )

        logging.info(f"Conversation Created: {response.name}")

        return response

    def complete_conversation(
        self,
        conversation_id
        ) -> types.Conversation:
        """Completes the specified conversation.
            Finished conversations are purged from the database after 30 days.

        Args:
            conversation_id (str)
                The conversation id that needs to be completed.

        Returns:
            Response for the completed conversation.
            type: google.cloud.dialogflow_v2beta1.types.Conversation
        """

        client = services.conversations.ConversationsClient(
            credentials = self.creds
            )
        conversation =  client.complete_conversation(
            name = conversation_id
            )
        logging.info(f"Closed Conversation {conversation_id}")

        return conversation

    def create_participant(
        self,
        conversation_id,
        role) -> types.Participant:
        """Creates a participant in a given conversation.

        Args:
            conversation_id (str):
                Required. The id of the conversation to which to
                add the participant.
            role (str):
                The role of the added participant.
                Allowed Values: HUMAN_AGENT, AUTOMATED_AGENT, END_USER

        Returns:
            Response for the created participant.
            type: google.cloud.dialogflow_v2beta1.types.Participant
        """

        client = services.participants.ParticipantsClient(
            credentials=self.creds
            )

        response = client.create_participant(
            parent = conversation_id,
            participant = {"role": role},
            timeout=600
            )

        logging.info(f"Participant created for the role {role}")
        logging.info(f"Participant Name: {response.name}")

        return response

    def list_participants(
        self,
        conversation_id: str
    ):
        """list all the participants added to a coversation
        Args:
        conversation_id (str):
                Required. The conversation to list all participants
                from. Format:
                ``projects/<Project ID>/locations/<Location ID>/conversations
                /<Conversation ID>``.
        Returns:
            A dict object with the participant information:
            {participant role: participant name}
        """
        client = services.participants.ParticipantsClient(
            credentials=self.creds
            )
        response = client.list_participants(parent = conversation_id)

        participant_dict = {}
        for participant in response:
            participant_dict[participant.role.name] = participant.name

        return participant_dict

    def get_participant(
        self,
        participant_id: str
    ) -> types.Participant:
        """Returns the Participant object for the given participant name
        Args:
            participant_id: str
                The participant id for the requested object.
                ``projects/<Project ID>/locations/<Location ID>/
                conversations/<Conv. ID>/participants/<Participant ID>``.
        """
        client = services.participants.ParticipantsClient(
            credentials = self.creds
            )

        request = types.GetParticipantRequest(
                    name=participant_id,
                )

        response = client.get_participant(request=request)

        return response

    def update_participant(
        self,
        participant_id: str = None,
        obj: types.Participant = None,
        **kwargs
    ) -> types.Participant:
        """
        Updates the participant object based on the provided args.
        Args:
            participant_id: str
                Full id of the participant to be updated.
            obj:
                The participant object to be updated.
        """
        if obj:
            participant = obj
            participant.name = participant_id

        else:
            if not participant_id:
                raise ValueError(
                    "Atleast one of participant object or participant "
                    "name is required to make the update."
                )

            participant = self.get_participant(
                participant_id = participant_id
                )

        # set participant attributes from kwargs
        for key, value in kwargs.items():
            setattr(participant, key, value)

        paths = kwargs.keys()
        mask = field_mask_pb2.FieldMask(paths=paths)

        client = services.participants.ParticipantsClient(
            credentials = self.creds
            )

        request = types.UpdateParticipantRequest()

        request.participant = participant
        request.update_mask = mask

        response = client.update_participant(request = request)

        return response

    def analyze_content_text(
        self,
        text,
        participant_id,
        language_code = "en-US"
        ) -> types.AnalyzeContentResponse:
        """Analyze text message content from a participant.

        Args:
            text (str):
                The input text string that needs to be analyzed.
            participant (str):
                The participant id of the participant for the
                corresponding text utterance being analyzed.
            language_code (str):
                Language of the inputted text.

        Returns:
            Response message from the analyze content.
            type: google.cloud.dialogflow.v2beta1.Participants.AnalyzeContent
        """

        client = services.participants.ParticipantsClient(
            credentials= self.creds
            )
        text_input = {"text": text, "language_code": language_code}

        logging.info("Running analyze content...")
        response = client.analyze_content(
            participant = participant_id,
            text_input = text_input
            )

        return response



"""A set of Utility methods to check resources stats on DFCX Agents."""

# Copyright 2023 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging
import time
from typing import Dict, List
import pandas as pd

from dfcx_scrapi.core import scrapi_base
from dfcx_scrapi.agent_extract import agents

# logging config
logging.basicConfig(
    level=logging.INFO,
    format="%(asctime)s %(levelname)-8s %(message)s",
    datefmt="%Y-%m-%d %H:%M:%S",
)

class AgentCheckerUtil(scrapi_base.ScrapiBase):
    """Utility class for checking DFCX Agents."""

    def __init__(
        self,
        agent_id: str,
        gcs_bucket_uri: str,
        creds_path: str = None,
        creds_dict: Dict = None,
        creds=None,
        scope=False,
    ):
        super().__init__(
            creds_path=creds_path,
            creds_dict=creds_dict,
            creds=creds,
            scope=scope,
        )
        self.agent_id = agent_id
        self.special_pages = [
            "End Session",
            "End Flow",
            "Start Page",
            "Current Page",
            "Previous Page",
        ]

        startup_time = time.time()
        self.extract = agents.Agents(agent_id)
        processing_time = time.time()
        logging.debug(f"STARTUP: {processing_time - startup_time}")

        self.data = self.extract.process_agent(agent_id, gcs_bucket_uri)
        logging.debug(f"TOTAL PROCESSING: {time.time() - processing_time}")

        self.active_intents_df = self.active_intents_to_dataframe()

    def _filter_special_pages(self, page: str, filter_special_pages: bool):
        """Recursion helper to check for special page match."""
        if filter_special_pages and page in self.special_pages:
            return True

        return False

    def _recurse_edges(self, edges: Dict[str, List[str]], page: str,
                      visited: set, depth: int, max_depth: int,
                      filter_special_pages: bool):
        """Recursion method used to traverse the agent graph for page data.

        Args:
          edges: The set of graph edges collected from the agent.
          page: The current Page Display Name
          visited: A set of visited Page nodes
          depth: The current recursion depth
          max_depth: The max recursion depth
          filter_special_pages: Will discard all self.special_pages from output
            if set to False.
        """
        if depth == max_depth:
            return visited

        if page in edges:
            for inner_page in edges[page]:
                if self._filter_special_pages(inner_page, filter_special_pages):
                    return visited

                if inner_page not in visited:
                    visited.add(inner_page)
                    visited = self._recurse_edges(
                        edges, inner_page, visited, depth+1, max_depth,
                        filter_special_pages)

        return visited

    def _mark_unreachable_pages(self, df: pd.DataFrame) -> pd.DataFrame:
        """Mark dataframe rows True if the page is unreachable in graph."""
        for idx, row in df.iterrows():
            for page in self.data.unreachable_pages[row["flow"]]:
                if row["page"] == page:
                    df.loc[idx, "unreachable"] = True

        return df

    def get_reachable_pages(
            self,
            flow_display_name: str,
            page_display_name: str = "Start Page",
            max_depth: int = 1,
            filter_special_pages = True) -> List[str]:
        """Get all pages in the graph that are reachable via transition routes,
        starting from a given Flow and Page.

        Args:
          flow_display_name: The display name of the flow
          page_display_name: The display name of the page. Defaults to
            "Start Page"
          max_depth: The max recursion depth to search the graph from the
            provided starting point. For example, a max_depth of 2 would produce
            all reachable Pages that are 2 transition routes away from the
            starting Flow/Page. Defaults to 1.
          filter_special_pages: Will filter out all self.special_pages. Defaults
            to True.
          """
        if page_display_name in ["START", "START_PAGE", "Start", "Start Page"]:
            page_display_name = "Start Page"
            page_display_name = f"{flow_display_name}: {page_display_name}"

        visited = self._recurse_edges(
            self.data.graph.edges, page_display_name, set(), 0, max_depth,
            filter_special_pages)

        return list(visited)

    def active_intents_to_dataframe(self) -> pd.DataFrame:
        """Gets all intents referenced in the agent, across all flows and pages,
        and produces a dataframe listing which flows/pages reference each
        intent.

        Returns:
            A dataframe with columns
            intent - the intent display name
            flow - the Flow Display Name where the intent resides
            page - the Page Display Name where the intent resides
            unreachable - Denotes whether the Flow/Page/Intent combination is
              unreachable in the graph.
        """
        df = pd.DataFrame({
            "intent": pd.Series(dtype="str"),
            "flow": pd.Series(dtype="str"),
            "page": pd.Series(dtype="str"),
            "unreachable": pd.Series(dtype="bool")
            })

        # Loop over active_intents, create temp dataframe, then concat with the
        # main dataframe to build out the complete Flow/Page/Intent dataset.
        for flow in self.data.active_intents:
            for pair in self.data.active_intents[flow]:
                intent = pair[0]
                page = pair[1]
                temp = pd.DataFrame({
                    "intent": [intent],
                    "flow": [flow],
                    "page": [page],
                    "unreachable": [False]})
                df = pd.concat([df, temp])

        df = df.reset_index(drop=True)

        # Finally, determine what rows are unreachable.
        self.active_intents_df = self._mark_unreachable_pages(df)

        return self.active_intents_df

    def get_unused_intents(self) -> List:
        """Get all unused Intents across the agent."""
        if self.active_intents_df.empty:
            self.active_intents_df = self.active_intents_to_dataframe()
        active_intents_set = set(self.active_intents_df.intent.to_list())
        all_intents_set = set(self.data.intents_map.keys())

        return list(all_intents_set.difference(active_intents_set))

    def get_unreachable_intents(self) -> pd.DataFrame:
        """Get all unreachable Intents across the agent.

        An Intent is unreachable if it resides on a page that is also
        unreachable.
        """
        if self.active_intents_df.empty:
            self.active_intents_df = self.active_intents_to_dataframe()

        return self.active_intents_df[self.active_intents_df["unreachable"]]



"""Test Case processing methods and functions."""

# Copyright 2023 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import json
import os

from typing import Dict, List, Any

from dfcx_scrapi.agent_extract import common
from dfcx_scrapi.agent_extract import types


class TestCases:
    """Test Case processing methods and functions."""

    def __init__(self):
        self.common = common.Common()

    @staticmethod
    def build_test_case_path_list(agent_local_path: str):
        """Builds a list of files, each representing a test case."""
        root_dir = agent_local_path + "/testCases"

        test_case_paths = []

        for test_case in os.listdir(root_dir):
            end = test_case.split(".")[-1]
            if end == "json":
                test_case_path = f"{root_dir}/{test_case}"
                test_case_paths.append(test_case_path)

        return test_case_paths

    @staticmethod
    def get_test_case_intent_phrase_pair(
        tc: types.TestCase) -> List[Dict[str, str]]:
        """Parse Test Case and return a list of intents in use.

        This method will produce a List of Dicts where the contents of each
        dict is the Training Phrase and associated Triggered Intent as listed
        in the Test Case Conversation Turn. This information is used to compare
        the User Input training phrase with the actual training phrases that
        exist in the Intent resource.

        The dict format is as follows:
            {
                training_phrase: <training_phrase>,
                intent: <intent_display_name>
            }
        """
        intent_data = []

        if tc.conversation_turns:
            for turn in tc.conversation_turns:
                user = turn["userInput"]
                agent = turn["virtualAgentOutput"]
                intent = agent.get("triggeredIntent", None)
                phrase = user.get("input", None)

                text = phrase.get("text", None)

                if text:
                    text = text["text"]

                if intent and text:
                    intent_data.append(
                        {
                            "user_utterance": text,
                            "intent": intent["name"],
                            "status": "valid",
                            "training_phrases": [],
                        }
                    )

        return intent_data

    @staticmethod
    def get_test_case_intent_data(agent_local_path: str):
        """Collect all Intent Files and Training Phrases for Test Case."""
        intents_path = agent_local_path + "/intents"

        intent_paths = []

        for intent_dir in os.listdir(intents_path):
            intent_dir_path = f"{intents_path}/{intent_dir}"
            intent_paths.append(
                {"intent": intent_dir, "file_path": intent_dir_path}
            )

        return intent_paths

    @staticmethod
    def flatten_tp_data(tp_data: List[Any]):
        """Flatten the Training Phrase proto to a list of strings."""
        cleaned_tps = []

        for tp in tp_data["trainingPhrases"]:
            parts_list = [part["text"].lower() for part in tp["parts"]]
            cleaned_tps.append("".join(parts_list))

        return cleaned_tps

    def gather_intent_tps(self, tc: types.TestCase):
        """Collect all TPs associated with Intent data in Test Case."""
        tc.associated_intent_data = {}

        for i, pair in enumerate(tc.intent_data):
            intent_dir = tc.agent_path + "/intents/" + pair["intent"]

            try:
                if "trainingPhrases" in os.listdir(intent_dir):
                    training_phrases_path = intent_dir + "/trainingPhrases"

                    for lang_file in os.listdir(training_phrases_path):
                        # lang_code = lang_file.split(".")[0]
                        lang_code_path = f"{training_phrases_path}/{lang_file}"

                        with open(
                            lang_code_path, "r", encoding="UTF-8"
                        ) as tp_file:
                            tp_data = json.load(tp_file)
                            cleaned_tps = self.flatten_tp_data(tp_data)

                            tp_file.close()

                        tc.intent_data[i]["training_phrases"].extend(
                            cleaned_tps
                        )
                        tc.associated_intent_data[pair["intent"]] = cleaned_tps

            except FileNotFoundError:
                tc.intent_data[i]["status"] = "invalid_intent"
                tc.has_invalid_intent = True
                continue

        return tc

    def process_test_case(self, tc: types.TestCase, stats: types.AgentData):
        """Process a single Test Case file."""

        with open(tc.dir_path, "r", encoding="UTF-8") as tc_file:
            tc.data = json.load(tc_file)
            tc.resource_id = tc.data.get("name", None)
            tc.display_name = tc.data.get("displayName", None)
            tc.tags = tc.data.get("tags", None)
            tc.conversation_turns = tc.data.get(
                "testCaseConversationTurns", None
            )
            tc.test_config = tc.data.get("testConfig", None)

            full_tc_id = f"{stats.agent_id}/testCases/{tc.resource_id}"
            tc.data["name"] = full_tc_id
            stats.test_cases.append(tc.data)

            tc_file.close()

        return stats

    def process_test_cases_directory(
            self, agent_local_path: str, stats: types.AgentData):
        """Processing the test cases dir in the JSON package structure."""
        test_case_paths = self.build_test_case_path_list(agent_local_path)
        stats.total_test_cases = len(test_case_paths)

        for test_case in test_case_paths:
            tc = types.TestCase()
            tc.dir_path = test_case
            tc.agent_path = agent_local_path
            stats = self.process_test_case(tc, stats)

        return stats



"""Agent processing methods and functions."""

# Copyright 2023 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging
import time
import os
import shutil
from typing import Dict

from dfcx_scrapi.core import agents
from dfcx_scrapi.core import operations
from dfcx_scrapi.core import scrapi_base
from dfcx_scrapi.agent_extract import graph
from dfcx_scrapi.agent_extract import flows
from dfcx_scrapi.agent_extract import intents
from dfcx_scrapi.agent_extract import entity_types
from dfcx_scrapi.agent_extract import test_cases
from dfcx_scrapi.agent_extract import webhooks
from dfcx_scrapi.agent_extract import gcs_utils
from dfcx_scrapi.agent_extract import types

# logging config
logging.basicConfig(
    level=logging.INFO,
    format="%(asctime)s %(levelname)-8s %(message)s",
    datefmt="%Y-%m-%d %H:%M:%S",
)

class Agents(scrapi_base.ScrapiBase):
    """Agent Metadata methods and functions."""
    def __init__(
        self,
        agent_id: str,
        lang_code: str = "en",
        creds_path: str = None,
        creds_dict: Dict = None,
        creds=None,
        scope=False
    ):
        super().__init__(
            creds_path=creds_path,
            creds_dict=creds_dict,
            creds=creds,
            scope=scope,
        )
        self.agent_id = agent_id
        self.lang_code = lang_code
        self._core_agents = agents.Agents(creds=creds)
        self.gcs = gcs_utils.GcsUtils()
        self.flows = flows.Flows()
        self.intents = intents.Intents()
        self.etypes = entity_types.EntityTypes()
        self.webhooks = webhooks.Webhooks()
        self.tcs = test_cases.TestCases()
        self.ops = operations.Operations()

    @staticmethod
    def prep_local_dir(agent_local_path: str):
        """Prepare the local directory for agent zip file."""
        if os.path.isdir(agent_local_path):
            logging.info("Cleaning up old directory...")
            shutil.rmtree(agent_local_path)
            logging.info(f"Making temp directory: {agent_local_path}")
            os.mkdir(agent_local_path)
        else:
            os.mkdir(agent_local_path)

    def await_lro(self, lro: str):
        """Wait for long running operation to complete."""
        try:
            i = 0
            while not self.ops.get_lro(lro).done:
                time.sleep(1)
                i += 1
                if i == 20:
                    break

        except UserWarning:
            print("LRO Failed.")

        return True

    def export_agent(self, agent_id: str, gcs_bucket_uri: str,
                      environment_display_name: str = None):
        """Handle the agent export, LRO and logging."""
        export_start = time.time()
        logging.info("Exporting agent...")
        lro = self._core_agents.export_agent(
            agent_id=agent_id,gcs_bucket_uri=gcs_bucket_uri, data_format="JSON",
            environment_display_name=environment_display_name)


        self.await_lro(lro)
        logging.info("Export Complete.")
        logging.debug(f"EXPORT: {time.time() - export_start}")

    def download_and_extract(self, agent_local_path: str, gcs_bucket_uri: str):
        """Handle download from GCS and extracting ZIP file."""
        if not os.path.exists(agent_local_path):
            os.makedirs(agent_local_path)

        download_start = time.time()
        logging.info("Downloading agent file from GCS Bucket...")
        agent_file = self.gcs.download_gcs(
            gcs_path=gcs_bucket_uri, local_path=agent_local_path)
        logging.info("Download complete.")
        logging.debug(f"DOWNLOAD: {time.time() - download_start}")

        self.gcs.unzip(agent_file, agent_local_path)


    def process_agent(self, agent_id: str, gcs_bucket_uri: str,
                      environment_display_name: str = None):
        """Process the specified Agent for offline data gathering."""
        agent_local_path = "/tmp/agent"
        self.prep_local_dir(agent_local_path)
        self.export_agent(agent_id, gcs_bucket_uri, environment_display_name)
        self.download_and_extract(agent_local_path, gcs_bucket_uri)

        logging.info("Processing Agent...")
        data = types.AgentData()
        data.graph = graph.Graph()
        data.lang_code = self.lang_code
        data.agent_id = agent_id
        data = self.flows.process_flows_directory(agent_local_path, data)
        data = self.intents.process_intents_directory(agent_local_path, data)
        data = self.etypes.process_entity_types_directory(
            agent_local_path, data)
        data = self.webhooks.process_webhooks_directory(agent_local_path, data)
        data = self.tcs.process_test_cases_directory(agent_local_path, data)
        logging.info("Processing Complete.")

        return data



"""Intent processing methods and functions."""

# Copyright 2023 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import json
import os

from dfcx_scrapi.agent_extract import common
from dfcx_scrapi.agent_extract import types


class Intents:
    """Intent processing methods and functions."""

    def __init__(self):
        self.common = common.Common()

    @staticmethod
    def parse_lang_code(lang_code_path: str) -> str:
        """Extract the language_code from the given file path."""

        first_parse = lang_code_path.split("/")[-1]
        lang_code = first_parse.split(".")[0]

        return lang_code

    @staticmethod
    def build_lang_code_paths(intent: types.Intent):
        """Builds dict of lang codes and file locations.

        The language_codes and paths for each file are stored in a dictionary
        inside of the Intent dataclass. This dict is access later to process
        each file and provide reporting based on each language code.
        """
        root_dir = intent.dir_path + "/trainingPhrases"

        for lang_file in os.listdir(root_dir):
            lang_code = lang_file.split(".")[0]
            lang_code_path = f"{root_dir}/{lang_file}"
            intent.training_phrases[lang_code] = {"file_path": lang_code_path}

    @staticmethod
    def build_intent_path_list(agent_local_path: str):
        """Builds a list of dirs, each representing an Intent directory.

        Ex: /path/to/agent/intents/<intent_dir>

        This dir path can be used to find the next level of information
        in the directory by appending the appropriate next dir structures like:
        - <intent_name>.json, for the Intent object metadata
        - /trainingPhrases, for the Training Phrases dir
        """
        root_dir = agent_local_path + "/intents"

        intent_paths = []

        for intent_dir in os.listdir(root_dir):
            intent_dir_path = f"{root_dir}/{intent_dir}"
            intent_paths.append(intent_dir_path)

        return intent_paths

    def process_intent_metadata(
            self, intent: types.Intent):
        """Process the metadata file for a single Intent."""
        intent.metadata_file = f"{intent.dir_path}/{intent.display_name}.json"

        try:
            with open(intent.metadata_file, "r", encoding="UTF-8") as meta_file:
                intent.data = json.load(meta_file)
                intent.resource_id = intent.data.get("name", None)
                intent.labels = intent.data.get("labels", None)
                intent.description = intent.data.get("description", None)
                intent.parameters = intent.data.get("parameters", None)

                meta_file.close()

        except FileNotFoundError:
            pass

    def process_language_codes(
            self, intent: types.Intent, stats: types.AgentData):
        """Process all training phrase lang_code files."""

        for lang_code in intent.training_phrases:
            tp_file = intent.training_phrases[lang_code]["file_path"]

            if not self.common.check_lang_code(lang_code, stats):
                continue

            with open(tp_file, "r", encoding="UTF-8") as tps:
                data = json.load(tps)
                data["name"] = f"{stats.agent_id}/intents/{intent.resource_id}"
                data["display_name"] = intent.display_name
                data["labels"] = intent.labels
                data["description"] = intent.description
                data["parameters"] = intent.parameters
                stats.intents.append(data)
                stats.total_training_phrases += len(data["trainingPhrases"])

                tps.close()

        return stats

    def process_training_phrases(
            self, intent: types.Intent, stats: types.AgentData):
        """Process the Training Phrase dir for a single Intent."""
        if "trainingPhrases" in os.listdir(intent.dir_path):
            self.build_lang_code_paths(intent)
            stats = self.process_language_codes(intent, stats)

        return stats

    def process_intent(self, intent: types.Intent, stats: types.AgentData):
        """Process a single Intent directory and associated files."""
        intent.display_name = self.common.parse_filepath(
            intent.dir_path, "intent")
        intent.display_name = self.common.clean_display_name(
            intent.display_name)

        self.process_intent_metadata(intent)
        stats = self.process_training_phrases(intent, stats)
        stats.total_intents += 1

        return stats

    def process_intents_directory(
            self, agent_local_path: str, stats: types.AgentData):
        """Processing the top level Intents Dir in the JSON Package structure.

        The following files/dirs exist under the `intents` dir:
        - <intent_display_name> Directory
          - trainingPhrases
            - <language-code>.json
          - <intent_display_name> Object

        In Dialogflow CX, the Training Phrases of each Intent are stored in
        individual .json files by language code under each Intent Display
        Name. In this method, we will process all Intent dirs, including the
        training phrase files and metadata objects for each Intent.
        """
        # Create a list of all Intent paths to iter through
        intent_paths = self.build_intent_path_list(agent_local_path)
        stats.intents = []

        for intent_path in intent_paths:
            intent = types.Intent()
            intent.dir_path = intent_path

            stats = self.process_intent(intent, stats)
            full_intent_id = f"{stats.agent_id}/intents/{intent.resource_id}"
            stats.intents_map[intent.display_name] = full_intent_id

        return stats



"""Entity Type processing methods and functions."""

# Copyright 2023 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import json
import os

from typing import Dict

from dfcx_scrapi.agent_extract import common
from dfcx_scrapi.agent_extract import types

class EntityTypes:
    """Entity Type processing methods and functions."""

    def __init__(self):
        self.common = common.Common()

    @staticmethod
    def build_entity_type_path_list(agent_local_path: str):
        """Builds a list of dirs, each representing an Entity Type directory.

        Ex: /path/to/agent/entityTypes/<entity_type_dir>

        This dir path can then be used to find the next level of information
        in the directory by appending the appropriate next dir structures like:
        - <entity_type_name>.json, for the Entity Type object
        - /entities, for the Entities dir
        """
        root_dir = agent_local_path + "/entityTypes"

        entity_type_paths = []

        for entity_type_dir in os.listdir(root_dir):
            entity_type_dir_path = f"{root_dir}/{entity_type_dir}"
            entity_type_paths.append(entity_type_dir_path)

        return entity_type_paths

    @staticmethod
    def build_lang_code_paths(etype: types.EntityType):
        """Builds dict of lang codes and file locations.

        The language_codes and paths for each file are stored in a dictionary
        inside of the Entity Type dataclass. This dict is accessed later to
        lint each file and provide reporting based on each language code.
        """
        root_dir = etype.dir_path + "/entities"

        for lang_file in os.listdir(root_dir):
            lang_code = lang_file.split(".")[0]
            lang_code_path = f"{root_dir}/{lang_file}"
            etype.entities[lang_code] = {"file_path": lang_code_path}

    @staticmethod
    def build_excluded_phrases_path(etype: types.EntityType, lang_code: str):
        """Builds a dict of excluded phrases and file locations."""
        root_dir = etype.dir_path + "/excludedPhrases"
        lang_code_path = f"{root_dir}/{lang_code}.json"

        return lang_code_path

    @staticmethod
    def process_entity_type_metadata(etype: types.EntityType):
        """Extract metadata for Entity Type for later processing."""
        metadata_file = etype.dir_path + f"/{etype.display_name}.json"

        with open(metadata_file, "r", encoding="UTF-8") as etype_file:
            etype.data = json.load(etype_file)
            etype.resource_id = etype.data.get("name", None)
            etype.kind = etype.data.get("kind", None)
            etype.auto_expansion = etype.data.get("autoExpansionMode", None)
            etype.fuzzy_extraction = etype.data.get(
                "enableFuzzyExtraction", False)

            etype_file.close()

    def process_excluded_phrases_language_codes(
            self, data: Dict[str, str], lang_code_path: str):
        """Process all ecluded phrases lang_code files."""
        with open(lang_code_path, "r", encoding="UTF-8") as ent_file:
            new_data = json.load(ent_file)
            data["excluded_phrases"] = new_data.get("excludedPhrases", None)

        return data

    def process_excluded_phrases(self, etype: types.EntityType, lang_code: str,
                                 data: Dict[str, str]):
        """Process the excluded phrases if they exist."""
        if "excludedPhrases" in os.listdir(etype.dir_path):
            lang_code_path = self.build_excluded_phrases_path(etype, lang_code)
            data = self.process_excluded_phrases_language_codes(
                data, lang_code_path)

        return data

    def process_language_codes(
            self, etype: types.EntityType, stats: types.AgentData):
        """Process all Entity Type lang_code files."""
        for lang_code in etype.entities:
            ent_file_path = etype.entities[lang_code]["file_path"]

            if not self.common.check_lang_code(lang_code, stats):
                continue

            with open(ent_file_path, "r", encoding="UTF-8") as ent_file:
                data = json.load(ent_file)
                data["name"] = f"{stats.agent_id}/entityTypes/"\
                    f"{etype.resource_id}"
                data["display_name"] = etype.display_name
                data["kind"] = etype.kind
                data["entities"] = data.get("entities", None)
                data = self.process_excluded_phrases(etype, lang_code, data)
                stats.entity_types.append(data)

                ent_file.close()

        return stats

    def process_entities(self, etype: types.EntityType, stats: types.AgentData):
        """Process the Entity files inside of an Entity Type."""
        if "entities" in os.listdir(etype.dir_path):
            self.build_lang_code_paths(etype)
            stats = self.process_language_codes(etype, stats)

        return stats

    def process_entity_type(
            self, etype: types.EntityType, stats: types.AgentData):
        """Process a Single Entity Type dir and all subdirectories."""

        etype.display_name = self.common.parse_filepath(
            etype.dir_path, "entity_type")
        etype.display_name = self.common.clean_display_name(etype.display_name)

        self.process_entity_type_metadata(etype)
        stats = self.process_entities(etype, stats)
        stats.total_entity_types += 1

        return stats

    def process_entity_types_directory(
            self, agent_local_path: str, stats: types.AgentData):
        """Processing the Entity Types dir in the JSON Package structure."""
        # Create a list of all Entity Type paths to iter through
        entity_type_paths = self.build_entity_type_path_list(agent_local_path)

        for entity_type_path in entity_type_paths:
            etype = types.EntityType()
            etype.dir_path = entity_type_path

            stats = self.process_entity_type(etype, stats)
            full_etype_id = f"{stats.agent_id}/entityTypes/{etype.resource_id}"
            stats.entity_types_map[etype.display_name] = full_etype_id

        return stats



"""Utility class for managing graph structure."""

# Copyright 2023 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import collections

class Graph:
    """Utility class for manaing graph structure."""

    def __init__(self):
        self.nodes = set()
        self.edges = collections.defaultdict(list)
        self.used_nodes = set()

    def add_node(self, node):
        """Add node to set of all nodes, regardless of use in graph."""
        self.nodes.add(node)

    def add_edge(self, node1, node2):
        self.edges[node1].append(node2)

    def add_used_node(self, node):
        """Add node to set of active in use nodes for the graph."""
        self.used_nodes.add(node)

    def remove_node(self, node):
        self.nodes.remove(node)

    def remove_edge(self, node1, node2):
        self.edges[node1].remove(node2)

    def __str__(self):
        return f"Graph({self.nodes}, {self.edges})"



"""A set of builder methods to create CX proto resource objects"""

# Copyright 2023 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging
from typing import List, Union

from google.cloud.dialogflowcx_v3beta1.types import TransitionRoute
from google.cloud.dialogflowcx_v3beta1.types import EventHandler

# logging config
logging.basicConfig(
    level=logging.INFO,
    format="%(asctime)s %(levelname)-8s %(message)s",
    datefmt="%Y-%m-%d %H:%M:%S",
)


class BuildersCommon:
    """Base class for other Builder classes"""

    _proto_type = None
    _proto_type_str = "None"


    def __init__(self, obj=None):
        self.proto_obj = None
        if obj:
            self.load_proto_obj(obj)


    def _check_proto_obj_attr_exist(self):
        """Check if the proto_obj exists otherwise raise an error."""
        if not self.proto_obj:
            raise ValueError(
                "There is no proto_obj!"
                "\nUse `create_new_proto_obj` or `load_proto_obj` to continue."
            )
        elif not isinstance(self.proto_obj, self._proto_type):  # pylint: disable=W1116
            raise ValueError(
                f"proto_obj is not {self._proto_type_str} type."
                "\nPlease create or load the correct type to continue."
            )


    def load_proto_obj(self, obj, overwrite: bool = False):
        """Load an existing object to proto_obj for further uses.

        Args:
          obj (proto object):
            An existing proto object.
          overwrite (bool)
            Overwrite the new proto_obj if proto_obj already
            contains some object.

        Returns:
          An object stored in proto_obj
        """
        if not isinstance(obj, self._proto_type):  # pylint: disable=W1116
            raise ValueError(
                "The object you're trying to load"
                f" is not {self._proto_type_str}!"
            )
        if self.proto_obj and not overwrite:
            raise UserWarning(
                f"proto_obj already contains {self._proto_type_str}."
                " If you wish to overwrite it, pass overwrite as True."
            )

        if overwrite or not self.proto_obj:
            self.proto_obj = obj

        return self.proto_obj


    def _is_type_or_list_of_types(self, obj, type_, var_name: str = None):
        """Check whether the `obj` type is `type_` or
        is a list with elements of `type_` otherwise raise an error.

        Args:
            obj:
              The object to check
            type_:
              Type of `obj`
            var_name (str):
              The variable name to show in the error message.

        Raises:
            ValueError: If the `obj` type is not `type_` or a list of `type_`s.
        """
        default_error_msg = "Incorrect type!!"
        error_msg_map = {
            str: (
                f"`{var_name}` should be either a string or a list of strings."
            ),
            EventHandler: (
                f"`{var_name}` should be either a EventHandler"
                " or a list of EventHandlers."
            ),
            TransitionRoute: (
                f"`{var_name}` should be either a TransitionRoute"
                " or a list of TransitionRoutes."
            ),
        }

        if not(
            isinstance(obj, type_) or
            (isinstance(obj, list) and
             all(isinstance(item, type_) for item in obj))
        ):
            msg = error_msg_map.get(obj, default_error_msg)
            raise ValueError(msg)


    def _match_transition_route(
        self,
        transition_route: TransitionRoute,
        target_route: TransitionRoute = None,
        intent: str = None,
        condition: str = None
    ) -> bool:
        """Check if transition_route's intent and condition
        matches with the input.

        At least one of the `target_route`, `intent`, or `condition` should
        be specfied.

        Args:
            transition_route (TransitionRoute):
              The TransitionRoute that input should match with.
            taget_route (TransitionRoute):
              The target TransitionRoute that we want to match.
            intent (str):
              TransitionRoute's intent that we want to match.
            condition (str):
              TransitionRoute's condition that we want to match.

        Returns:
            True if the `transition_route` matched with the input.
        """
        # Type/Error checking
        if not isinstance(transition_route, TransitionRoute):
            raise ValueError(
                "`transition_route` should have the type TransitionRoute."
            )
        if not(target_route or intent or condition):
            raise ValueError(
                "At least one of `target_route`, `intent`, or `condition`"
                " must be specified."
            )

        # Check if the transition route matches
        is_match = False
        if target_route:
            is_match = self._check_transition_route_with_target_route(
                transition_route, target_route
            )
        if intent and condition:
            is_match = self._check_transition_route_with_intent_and_condition(
                transition_route, intent, condition
            )
        elif intent and not condition:
            is_match = self._check_transition_route_with_intent(
                transition_route, intent
            )
        elif not intent and condition:
            is_match = self._check_transition_route_with_condition(
                transition_route, condition
            )

        return is_match


    def _check_transition_route_with_target_route(
        self,
        transition_route: TransitionRoute,
        target_route: TransitionRoute
    ) -> bool:
        """Check if transition_route's intent and condition
        matches with the target_route's intent and condition.

        Args:
            transition_route (TransitionRoute):
              The TransitionRoute that input should match with.
            taget_route (TransitionRoute):
              The target TransitionRoute that we want to match.

        Returns:
            True if the `transition_route` matched with the input.
        """
        # Type/Error checking
        if not isinstance(target_route, TransitionRoute):
            raise ValueError("`target_route` should be a TransitionRoute.")

        if (
            transition_route.condition == target_route.condition and
            transition_route.intent == target_route.intent
        ):
            return True
        return False


    def _check_transition_route_with_intent_and_condition(
        self,
        transition_route: TransitionRoute,
        intent: str,
        condition: str
    ) -> bool:
        """Check if transition_route's intent and condition
        matches with the input.

        Args:
            transition_route (TransitionRoute):
              The TransitionRoute that input should match with.
            intent (str):
              TransitionRoute's intent that we want to match.
            condition (str):
              TransitionRoute's condition that we want to match.

        Returns:
            True if the `transition_route` matched with the input.
        """
        # Type/Error checking
        if not(isinstance(intent, str) and isinstance(condition, str)):
            raise ValueError("`intent` and `condition` should be a string.")

        if (
            transition_route.condition == condition and
            transition_route.intent == intent
        ):
            return True
        return False


    def _check_transition_route_with_intent(
        self,
        transition_route: TransitionRoute,
        intent: str
    ) -> bool:
        """Check if transition_route's intent matches with the input.

        Args:
            transition_route (TransitionRoute):
              The TransitionRoute that input should match with.
            intent (str):
              TransitionRoute's intent that we want to match.

        Returns:
            True if the `transition_route` matched with the input.
        """
        # Type/Error checking
        if not isinstance(intent, str):
            raise ValueError("`intent` should be a string.")

        if transition_route.intent == intent:
            return True
        return False


    def _check_transition_route_with_condition(
        self,
        transition_route: TransitionRoute,
        condition: str
    ) -> bool:
        """Check if transition_route's condition matches with the input.

        Args:
            transition_route (TransitionRoute):
              The TransitionRoute that input should match with.
            condition (str):
              TransitionRoute's condition that we want to match.

        Returns:
            True if the `transition_route` matched with the input.
        """
        # Type/Error checking
        if not isinstance(condition, str):
            raise ValueError("`condition` should be a string.")

        if transition_route.condition == condition:
            return True
        return False


    def _find_unmatched_event_handlers(
        self, event_handlers: Union[EventHandler, List[EventHandler]]
    ) -> List[EventHandler]:
        """Find the EventHandlers of proto_obj which is not present
        in the `event_handlers`

        Args:
          event_handlers (EventHandler | List[EventHandler]):
            A single or list of EventHandler to remove
              from the existing EventHandlers in proto_obj.

        Returns:
          A list of EventHandlers
        """
        # Type error checking
        self._is_type_or_list_of_types(
            event_handlers, EventHandler, "event_handlers"
        )

        if not isinstance(event_handlers, list):
            event_handlers = [event_handlers]

        return [
            eh
            for eh in self.proto_obj.event_handlers
            if eh not in event_handlers
        ]


    def _find_unmatched_event_handlers_by_name(
        self, event_names: Union[str, List[str]]
    ) -> List[EventHandler]:
        """Find the EventHandlers of proto_obj which their event names
        is not present in the `event_names`

        Args:
          event_names (str | List[str]):
            A single or list of EventHandler's event names corresponding
              to the EventHandler to remove from the existing
              EventHandlers in proto_obj.

        Returns:
          A list of EventHandlers
        """
        # Type error checking
        self._is_type_or_list_of_types(event_names, str, "event_names")

        if not isinstance(event_names, list):
            event_names = [event_names]

        return [
            eh
            for eh in self.proto_obj.event_handlers
            if eh.event not in event_names
        ]



"""A set of builder methods to create CX proto resource objects"""

# Copyright 2023 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging
from string import ascii_lowercase
from string import digits
from collections import defaultdict
from dataclasses import dataclass
from dataclasses import field
from typing import List, Dict, Union

from google.cloud.dialogflowcx_v3beta1.types import Intent
from dfcx_scrapi.builders.builders_common import BuildersCommon

# logging config
logging.basicConfig(
    level=logging.INFO,
    format="%(asctime)s %(levelname)-8s %(message)s",
    datefmt="%Y-%m-%d %H:%M:%S",
)


class IntentBuilder(BuildersCommon):
    """Base Class for CX Intent builder."""

    _proto_type = Intent
    _proto_type_str = "Intent"


    def __str__(self) -> str:
        """String representation of the proto_obj."""
        self._check_proto_obj_attr_exist()

        return (f"{self._show_basic_info()}"
            f"\n\n{self._show_parameters()}"
            f"\n\n{self._show_training_phrases()}")


    def _include_spaces_to_phrase(self, phrase: List[str], annots: List[str]):
        """Internal method to add spaces to the training phrase list and
        make related changes to the annotations list.

        Args:
          phrase (List[str]):
            A list of strings that represents the training phrase.
          annots (List[str]):
            A list of strings that represents
              parameter_id of each part in phrase.
        """
        chars_to_ignore_at_beginning = ["'", ",", ".", "?", "!"]
        i = 0
        while True:
            p_curr, a_curr = phrase[i], annots[i]
            try:
                p_next, a_next = phrase[i+1], annots[i+1]
            except IndexError:
                break

            if a_curr and a_next:
                phrase.insert(i+1, " ")
                annots.insert(i+1, "")
                i += 2
            elif a_curr and not a_next:
                flag = any(
                    ch
                    for ch in chars_to_ignore_at_beginning
                    if p_next.startswith(ch)
                )
                if not flag:
                    phrase[i+1] = " " + p_next
                i += 1
            elif not a_curr and a_next:
                phrase[i] = p_curr + " "
                i += 1
            elif not a_curr and not a_next:
                phrase[i] = p_curr + " " + p_next
                del phrase[i+1]
                del annots[i+1]


    def _label_constraints_check(self, key: str, value: str):
        """Check constraints for the label's key and value
        and raise an error if needed.

        Args:
          key (str):
            Label's key.
          value (str):
            Label's value.
        """
        allowed_chars = ascii_lowercase + digits + "-_"
        # TODO Add International characteres to allowed_chars
        allowed_char_error_msg = (
            "Key and Value can only contain lowercase letters,"
            " numeric characters, underscores and dashes."
        )
        if not(isinstance(key, str) and isinstance(value, str)):
            raise ValueError(
                "Key and value should be string."
            )
        if len(key) > 63 or len(value) > 63:
            raise ValueError(
                "Key and value can be no longer than 63 characters"
            )
        if key.startswith("sys-") and key not in ["sys-head", "sys-contextual"]:
            raise ValueError(
                "Prefix `sys-` is reserved for Dialogflow defined labels."
            )
        if key[0] not in ascii_lowercase:
            raise ValueError("Key must start with a lowercase letter.")
        for s in key:
            if s not in allowed_chars:
                raise ValueError(allowed_char_error_msg)
        for s in value:
            if s not in allowed_chars:
                raise ValueError(allowed_char_error_msg)


    def _show_basic_info(self) -> str:
        """String representation for the basic information of proto_obj."""
        self._check_proto_obj_attr_exist()

        labels = [
            str(key) if key == val else f"{key}: {val}"
            for key, val in self.proto_obj.labels.items()
        ]
        return (f"display_name: {self.proto_obj.display_name}"
            f"\ndescription: {self.proto_obj.description}"
            f"\npriority: {self.proto_obj.priority}"
            f"\nis_fallback: {self.proto_obj.is_fallback}"
            f"\nlabels: {labels}")


    def _show_parameters(self) -> str:
        """String representation for the parameters of proto_obj."""
        self._check_proto_obj_attr_exist()

        return "\n".join([
            (f"parameter_id: {str(param.id)}"
            f"\nentity_type: {str(param.entity_type)}"
            f"\n\tis_list: {bool(param.is_list)}"
            f"\n\tredact: {bool(param.redact)}")
            for param in self.proto_obj.parameters
        ])


    def _show_training_phrases(self, repeat_count: int = None) -> str:
        """String representation for the training phrases of proto_obj."""
        self._check_proto_obj_attr_exist()

        phrases = []
        for tp in self.proto_obj.training_phrases:
            if repeat_count and tp.repeat_count != repeat_count:
                continue
            phrase = "".join([part.text for part in tp.parts])
            annotations = {
                part.text: part.parameter_id
                for part in tp.parts
                if part.parameter_id
            }
            phrases.append(
                f"phrase: {phrase}"
                f"\nannotations: {str(annotations)}"
                f"\n\trepeat_count: {tp.repeat_count}"
            )

        return "\n".join(phrases)


    def show_intent(
        self, mode: str = "whole", repeat_count: int = None
    ) -> None:
        """Show the proto_obj information.

        Args:
          mode (str):
            Specifies what part of the intent to show.
              Options:
              ['basic', 'parameters', 'phrases' or 'training phrases', 'whole']
          repeat_count (int):
            Indicates how many times the training phrases
            was added to the intent.
        """
        self._check_proto_obj_attr_exist()

        self.parameter_checking()

        if mode == "basic":
            print(self._show_basic_info())
        elif mode == "parameters":
            print(self._show_parameters())
        elif mode in ["phrases", "training phrases"]:
            print(self._show_training_phrases(repeat_count=repeat_count))
        elif mode == "whole":
            print(self)
        else:
            raise ValueError(
                "mode should be in"
                " ['basic', 'parameters',"
                " 'phrases', 'training phrases', 'whole']"
            )


    def show_stats(self):
        """Provide some stats about the intent."""
        self._check_proto_obj_attr_exist()

        stats_instance = IntentStats(self.proto_obj)
        stats_instance.generate_stats()


    def parameter_checking(self, raise_error: bool = False) -> bool:
        """Check if the annotated parameters exist
        in the Parameter attribute of proto_obj.

        Args:
          raise_error (bool):
            A flag to whether raise an error. If False, it will log a warning.

        Returns:
          True if annotated parameters are the same as parameters in proto_obj
        """
        self._check_proto_obj_attr_exist()

        tp_params_set = set()
        for tp in self.proto_obj.training_phrases:
            for part in tp.parts:
                tp_params_set.add(part.parameter_id)
        # Remove the empty string for unannotated parts
        try:
            tp_params_set.remove("")
        except KeyError:
            pass

        # Get the parameters from proto_obj
        parameters_set = {param.id for param in self.proto_obj.parameters}

        # Check for not existing annotated parameters
        return_flag = True
        for tp_param in tp_params_set:
            if tp_param not in parameters_set:
                return_flag = False
                msg = (
                    f"parameter_id `{tp_param}` does not exist in parameters."
                    "\nPlease add it using add_parameter method to continue."
                )
                if raise_error:
                    raise UserWarning(msg)
                else:
                    logging.warning(msg)

        return bool(return_flag)


    def create_new_proto_obj(
        self,
        display_name: str,
        priority: int = 500000,
        is_fallback: bool = False,
        description: str = None,
        overwrite: bool = False,
    ) -> Intent:
        """Create a new Intent.

        Args:
          display_name (str):
            Required. The human-readable name of the
            intent, unique within the agent.
          priority (int):
            The priority of this intent. Higher numbers represent higher
            priorities.
            -  If the supplied value is unspecified or 0, the service
            translates the value to 500,000, which corresponds to the
            ``Normal`` priority in the console.
            -  If the supplied value is negative, the intent is ignored
            in runtime detect intent requests.
          is_fallback (bool):
            Indicates whether this is a fallback intent.
            Currently only default fallback intent is
            allowed in the agent, which is added upon agent
            creation.
            Adding training phrases to fallback intent is
            useful in the case of requests that are
            mistakenly matched, since training phrases
            assigned to fallback intents act as negative
            examples that triggers no-match event.
          description (str):
            Human readable description for better
            understanding an intent like its scope, content,
            result etc. Maximum character limit: 140
            characters.
          overwrite (bool)
            Overwrite the new proto_obj if proto_obj already
            contains an Intent.

        Returns:
          An Intent object stored in proto_obj
        """
        if self.proto_obj and not overwrite:
            raise UserWarning(
                "proto_obj already contains an Intent."
                " If you wish to overwrite it, pass overwrite as True."
            )

        if overwrite or not self.proto_obj:
            self.proto_obj = Intent(
                display_name=display_name,
                priority=priority,
                is_fallback=is_fallback,
                description=description
            )

        return self.proto_obj


    def add_training_phrase(
        self,
        phrase: Union[str, List[str]],
        annotations: List[str] = None,
        repeat_count: int = 1,
        include_spaces: bool = True,
    ) -> Intent:
        """Add a training phrase to proto_obj.

        Args:
          phrase (str | List[str]):
            The training phrase as a string without annotations or
            a list of strings that represents a single training phrase.
          annotations (List[str]):
            A list of strings that represents
              parameter_id of each part in phrase.
            Length of annotations list should be less than or equal to
              length of phrase list.
            If the length is less than length of phrase list it propagate
              the rest of the annotations automatically with no annotation.
          repeat_count (int):
            Indicates how many times this example was added to the intent.
          include_spaces (bool):
            For the phrases with annotations, indicates whether the function
              should include spaces between each part of training phrase.

        Example 1: phrase = "I want to check my balance"
        Example 2:
          phrase = ['I want to order a', 'pizza']
          annotations = ['', 'food_type']
          include_spaces = True
        Example 3:
          phrase = [
              'one way', ' ticket leaving ', 'January 1',
              ' to ', 'LAX', ' from ', 'CDG'
          ]
          annotations = [
              'flight_type', '', 'departure_date',
              '', 'arrival_city', '', 'departure_city'
          ]
          include_spaces = False
        Example 4:
          phrase = ["I'd like to buy a", 'one way', 'ticket']
          annotations = ['', 'flight_type']
          include_spaces = True

        Returns:
          An Intent object stored in proto_obj
        """
        self._check_proto_obj_attr_exist()

        # Add simple training phrase
        if isinstance(phrase, str):
            # Create the training phrase obj and add it to the others
            tp = Intent.TrainingPhrase(
                parts=[Intent.TrainingPhrase.Part(text=phrase)],
                repeat_count=repeat_count
            )
            self.proto_obj.training_phrases.append(tp)

            return self.proto_obj

        # Add annotated training phrase
        # Type / Error checking
        if not (isinstance(phrase, list) and isinstance(annotations, list)):
            raise ValueError(
                "Both phrase and annotations should be lists."
            )
        if not (
            all((isinstance(p, str) for p in phrase)) and
            all((isinstance(a, str) for a in annotations))
        ):
            raise ValueError(
                "Only strings allowed in phrase or annotations list."
            )
        if len(annotations) > len(phrase):
            raise IndexError(
                "Length of annotations list is more than phrase list!"
            )

        # Propagate the annotations list if needed
        if len(annotations) < len(phrase):
            annotations.extend([""] * (len(phrase) - len(annotations)))
        # Change the phrase and annotations lists to include spaces if needed
        if include_spaces:
            self._include_spaces_to_phrase(phrase, annotations)
        # Creating parts for the training phrase
        parts_list = []
        for text, parameter_id in zip(phrase, annotations):
            part = Intent.TrainingPhrase.Part(
                text=text, parameter_id=parameter_id
            )
            parts_list.append(part)

        # Create the training phrase obj and add it to the others
        tp = Intent.TrainingPhrase(
            parts=parts_list, repeat_count=repeat_count
        )
        self.proto_obj.training_phrases.append(tp)

        self.parameter_checking()

        return self.proto_obj


    def add_parameter(
        self,
        parameter_id: str,
        entity_type: str,
        is_list: bool = False,
        redact: bool = False,
    ) -> Intent:
        """Add a parameter to Parameter attribute of proto_obj.

        Args:
          parameter_id (str):
            Required. The unique identifier of the parameter.
          entity_type (str):
            Required. The entity type of the parameter. Format:
            ``projects/-/locations/-/agents/-/
              entityTypes/<System Entity Type ID>``
            for system entity types (for example,
            ``projects/-/locations/-/agents/-/entityTypes/sys.date``),
            or
            ``projects/<Project ID>/locations/<Location ID>/
              agents/<Agent ID>/entityTypes/<Entity Type ID>``
            for developer entity types.
          is_list (bool):
            Indicates whether the parameter represents a
            list of values.
          redact (bool):
            Indicates whether the parameter content should be redacted
            in log. If redaction is enabled, the parameter content will
            be replaced by parameter name during logging.

        Returns:
          An Intent object stored in proto_obj
        """
        self._check_proto_obj_attr_exist()

        if not (isinstance(parameter_id, str) or isinstance(entity_type, str)):
            raise ValueError(
                "parameter_id and entity_type should be string."
            )

        # Create the new parameter and add it to the proto_obj
        parameters = Intent.Parameter(
            id=parameter_id,
            entity_type=entity_type,
            is_list=is_list,
            redact=redact
        )
        self.proto_obj.parameters.append(parameters)

        return self.proto_obj


    def add_label(
        self, label: Union[Dict[str, str], str]
    ) -> Intent:
        """Add a label to proto_obj.

        Args:
          labels (Dict[str, str] | str):
            labels can be assigned as key:value like driver:account
            or strings like 'head intent'.
            Labels can contain lowercase letters, digits and
              the symbols '-' and'_'.
            International characters are allowed, including letters
              from unicase alphabets.
            Keys must start with a letter.
            Keys and values can be no longer than 63 characters and no more
              than 128 bytes.
            Prefix "sys-" is reserved for Dialogflow defined labels.
              Currently allowed Dialogflow defined labels include:
                - "sys-head" means the intent is a head intent.
                - "sys-contextual" means the intent is a contextual intent.

        Returns:
          An Intent object stored in proto_obj
        """
        self._check_proto_obj_attr_exist()

        if isinstance(label, str):
            self._label_constraints_check(key=label, value=label)
            self.proto_obj.labels.update({label: label})
        elif isinstance(label, dict):
            for key, val in label.items():
                self._label_constraints_check(key=key, value=val)
            self.proto_obj.labels.update(label)
        else:
            raise ValueError(
                "label should be either a string or a dictionary."
            )

        return self.proto_obj


    def remove_training_phrase(self, phrase: str) -> Intent:
        """Remove a training phrase from proto_obj.

        Args:
          phrase (str):
            The training phrase to remove from proto_obj.

        Returns:
          An Intent object stored in proto_obj
        """
        self._check_proto_obj_attr_exist()

        if not isinstance(phrase, str):
            raise ValueError("phrase should be a string.")

        for idx, tp in enumerate(self.proto_obj.training_phrases):
            # Construct the training phrase using parts
            the_phrase = "".join([part.text for part in tp.parts])
            if phrase == the_phrase:
                del self.proto_obj.training_phrases[idx]
                break

        return self.proto_obj


    def remove_parameter(self, parameter_id: str) -> Intent:
        """Remove a parameter from proto_obj.

        Args:
          parameter_id (str):
            The id of the parameter to remove from proto_obj.

        Returns:
          An Intent object stored in proto_obj
        """
        self._check_proto_obj_attr_exist()

        if not isinstance(parameter_id, str):
            raise ValueError("parameter_id should be a string.")

        for idx, param in enumerate(self.proto_obj.parameters):
            if parameter_id == param.id:
                del self.proto_obj.parameters[idx]
                break

        return self.proto_obj


    def remove_label(self, label: Union[Dict[str, str], str]) -> Intent:
        """Remove a single or multiple labels from proto_obj.

        Args:
          label (Dict[str, str] | str):
            A string or a dictionary of lables to remove.

        Returns:
          An Intent object stored in proto_obj
        """
        self._check_proto_obj_attr_exist()

        if isinstance(label, str):
            if self.proto_obj.labels.get(label) == label:
                self.proto_obj.labels.pop(label)
        elif isinstance(label, dict):
            for key, val in label.items():
                if not(isinstance(key, str) and isinstance(val, str)):
                    raise ValueError(
                        "Keys and values in label's dictionary"
                        " should be string."
                    )
                # Check if the keys and values in the `label`
                # are the same as labels in proto_obj
                if self.proto_obj.labels.get(key) == val:
                    self.proto_obj.labels.pop(key)
        else:
            raise ValueError(
                "labels should be either a string or a dictionary."
            )

        return self.proto_obj



@dataclass
class IntentStats():
    """A class for tracking the stats of CX Intent object."""
    intent_proto_obj: Intent

    # Training Phrases
    repeat_count_dict: Dict[int, int] = field(
        default_factory=lambda: defaultdict(int)
    )
    annotated_count: int = 0
    phrases_count: int = 0
    annotated_pct: int = 0
    uniques_count: int = 0
    uniques_pct: int = 0
    non_uniques_count: int = 0
    non_uniques_pct: int = 0

    # Parameters
    params_count: int = 0


    def calc_tps_stats(self):
        """Calculating stats related to training phrases."""
        for tp in self.intent_proto_obj.training_phrases:
            self.repeat_count_dict[tp.repeat_count] += 1

            for part in tp.parts:
                if part.parameter_id:
                    self.annotated_count += 1
                    break


        self.phrases_count = len(self.intent_proto_obj.training_phrases)
        self.uniques_count = self.repeat_count_dict[1]
        self.non_uniques_count = self.phrases_count - self.uniques_count

        annot_pct = 100 * (self.annotated_count / self.phrases_count)
        uniq_pct = 100 * (self.uniques_count / self.phrases_count)
        non_uniq_pct = 100 * (self.non_uniques_count / self.phrases_count)
        self.annotated_pct = round(annot_pct, 1)
        self.uniques_pct = round(uniq_pct, 1)
        self.non_uniques_pct = round(non_uniq_pct, 1)

    def create_tps_str(self) -> str:
        """String representation of stats related to training phrases."""
        phrases_str = f"# of training phrases: {self.phrases_count}"
        annotated_str = (
            "Annotated training phrases:"
            f" {self.annotated_pct}% ({self.annotated_count})"
        )

        uniques_str = (
            "Unique training phrases:"
            f" {self.uniques_pct}% ({self.uniques_count})"
        )
        non_uniques_str = (
            "Non-unique training phrases:"
            f" {self.non_uniques_pct}% ({self.non_uniques_count})"
        )

        repeat_count_srt = "\n\t".join([
            f"with repeat count {i}: {self.repeat_count_dict[i]}"
            for i in sorted(self.repeat_count_dict.keys())
        ])

        return (
            f"{phrases_str}\n{annotated_str}\n"
            f"\n{uniques_str}\n{non_uniques_str}\n\t{repeat_count_srt}"
        )


    def create_parameter_str(self) -> str:
        """String representation Intent's parameters.."""
        self.params_count = len(self.intent_proto_obj.parameters)
        return f"# of parameters: {self.params_count}"


    def generate_stats(self):
        """Generate stats for the Intent."""
        self.calc_tps_stats()

        tps_str = self.create_tps_str()
        params_str = self.create_parameter_str()

        out = f"{params_str}\n{tps_str}"
        print(out)
